# Copyright (c) 2018 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import re
import time
from unittest import mock

import six

from cinder import context
from cinder import exception
from cinder.objects import volume_attachment
from cinder.tests.unit.consistencygroup import fake_consistencygroup as fake_cg
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit.fake_volume import fake_volume_obj
from cinder.tests.unit.fake_volume import fake_volume_type_obj
from cinder.tests.unit import test
from cinder.volume.drivers.dell_emc import xtremio

typ2id = {'volumes': 'vol-id',
          'snapshots': 'vol-id',
          'initiators': 'initiator-id',
          'initiator-groups': 'ig-id',
          'lun-maps': 'mapping-id',
          'consistency-groups': 'cg-id',
          'consistency-group-volumes': 'cg-vol-id',
          }

xms_init = {'xms': {1: {'version': '4.2.0',
                        'sw-version': '4.2.0-30'}},
            'clusters': {1: {'name': 'brick1',
                             'sys-sw-version': "4.2.0-devel_ba23ee5381eeab73",
                             'ud-ssd-space': '8146708710',
                             'ud-ssd-space-in-use': '708710',
                             'vol-size': '29884416',
                             'chap-authentication-mode': 'disabled',
                             'chap-discovery-mode': 'disabled',
                             "index": 1,
                             },
                         },
            'target-groups': {'Default': {"index": 1, "name": "Default"},
                              },
            'iscsi-portals': {'10.205.68.5/16':
                              {"port-address":
                               "iqn.2008-05.com.xtremio:001e67939c34",
                               "ip-port": 3260,
                               "ip-addr": "10.205.68.5/16",
                               "name": "10.205.68.5/16",
                               "index": 1,
                               },
                              },
            'targets': {'X1-SC2-target1': {'index': 1, "name": "X1-SC2-fc1",
                                           "port-address":
                                           "21:00:00:24:ff:57:b2:36",
                                           'port-type': 'fc',
                                           'port-state': 'up',
                                           },
                        'X1-SC2-target2': {'index': 2, "name": "X1-SC2-fc2",
                                           "port-address":
                                           "21:00:00:24:ff:57:b2:55",
                                           'port-type': 'fc',
                                           'port-state': 'up',
                                           }
                        },
            'volumes': {},
            'initiator-groups': {},
            'initiators': {},
            'lun-maps': {},
            'consistency-groups': {},
            'consistency-group-volumes': {},
            }

xms_data = None

xms_filters = {
    'eq': lambda x, y: x == y,
    'ne': lambda x, y: x != y,
    'gt': lambda x, y: x > y,
    'ge': lambda x, y: x >= y,
    'lt': lambda x, y: x < y,
    'le': lambda x, y: x <= y,
}


def get_xms_obj_by_name(typ, name):
    for item in xms_data[typ].values():
        if 'name' in item and item['name'] == name:
            return item
    raise exception.NotFound()


def clean_xms_data():
    global xms_data
    xms_data = copy.deepcopy(xms_init)


def fix_data(data, object_type):
    d = {}
    for key, value in data.items():
        if 'name' in key:
            key = 'name'
        d[key] = value

    if object_type == 'lun-maps':
        d['lun'] = 1

        vol_idx = get_xms_obj_by_name('volumes', data['vol-id'])['index']
        ig_idx = get_xms_obj_by_name('initiator-groups',
                                     data['ig-id'])['index']

        d['name'] = '_'.join([six.text_type(vol_idx),
                              six.text_type(ig_idx),
                              '1'])

    d[typ2id[object_type]] = ["a91e8c81c2d14ae4865187ce4f866f8a",
                              d.get('name'),
                              len(xms_data.get(object_type, [])) + 1]
    d['index'] = len(xms_data[object_type]) + 1
    return d


def get_xms_obj_key(data):
    for key in data.keys():
        if 'name' in key:
            return key


def get_obj(typ, name, idx):
    if name:
        return {"content": get_xms_obj_by_name(typ, name)}
    elif idx:
        if idx not in xms_data.get(typ, {}):
            raise exception.NotFound()
        return {"content": xms_data[typ][idx]}


def xms_request(object_type='volumes', method='GET', data=None,
                name=None, idx=None, ver='v1'):
    if object_type == 'snapshots':
        object_type = 'volumes'

    try:
        res = xms_data[object_type]
    except KeyError:
        raise exception.VolumeDriverException
    if method == 'GET':
        if name or idx:
            return get_obj(object_type, name, idx)
        else:
            if data and data.get('full') == 1:
                filter_term = data.get('filter')
                if not filter_term:
                    entities = list(res.values())
                else:
                    field, oper, value = filter_term.split(':', 2)
                    comp = xms_filters[oper]
                    entities = [o for o in res.values()
                                if comp(o.get(field), value)]
                return {object_type: entities}
            else:
                return {object_type: [{"href": "/%s/%d" % (object_type,
                                                           obj['index']),
                                       "name": obj.get('name')}
                                      for obj in res.values()]}
    elif method == 'POST':
        data = fix_data(data, object_type)
        name_key = get_xms_obj_key(data)
        try:
            if name_key and get_xms_obj_by_name(object_type, data[name_key]):
                raise (exception
                       .VolumeBackendAPIException
                       ('Volume by this name already exists'))
        except exception.NotFound:
            pass
        data['index'] = len(xms_data[object_type]) + 1
        xms_data[object_type][data['index']] = data
        # find the name key
        if name_key:
            data['name'] = data[name_key]
        if object_type == 'lun-maps':
            data['ig-name'] = data['ig-id']

        return {"links": [{"href": "/%s/%d" %
                          (object_type, data[typ2id[object_type]][2])}]}
    elif method == 'DELETE':
        if object_type == 'consistency-group-volumes':
            data = [cgv for cgv in
                    xms_data['consistency-group-volumes'].values()
                    if cgv['vol-id'] == data['vol-id']
                    and cgv['cg-id'] == data['cg-id']][0]
        else:
            data = get_obj(object_type, name, idx)['content']
        if data:
            del xms_data[object_type][data['index']]
        else:
            raise exception.NotFound()
    elif method == 'PUT':
        obj = get_obj(object_type, name, idx)['content']
        data = fix_data(data, object_type)
        del data['index']
        obj.update(data)


def xms_bad_request(object_type='volumes', method='GET', data=None,
                    name=None, idx=None, ver='v1'):
    if method == 'GET':
        raise exception.NotFound()
    elif method == 'POST':
        raise exception.VolumeBackendAPIException('Failed to create ig')


def xms_failed_rename_snapshot_request(object_type='volumes',
                                       method='GET', data=None,
                                       name=None, idx=None, ver='v1'):
    if method == 'POST':
        xms_data['volumes'][27] = {}
        return {
            "links": [
                {
                    "href": "https://host/api/json/v2/types/snapshots/27",
                    "rel": "self"}]}
    elif method == 'PUT':
        raise exception.VolumeBackendAPIException(data='Failed to delete')
    elif method == 'DELETE':
        del xms_data['volumes'][27]


class D(dict):
    def update(self, *args, **kwargs):
        self.__dict__.update(*args, **kwargs)
        return dict.update(self, *args, **kwargs)


class CommonData(object):
    context = context.RequestContext('admin', 'fake', True)
    connector = {'ip': '10.0.0.2',
                 'initiator': 'iqn.1993-08.org.debian:01:222',
                 'wwpns': ["123456789012345", "123456789054321"],
                 'wwnns': ["223456789012345", "223456789054321"],
                 'host': 'fakehost',
                 }

    test_volume_type = fake_volume_type_obj(
        context=context
    )

    test_volume = fake_volume_obj(context,
                                  volume_type = test_volume_type,
                                  name='vol1',
                                  volume_name='vol1',
                                  display_name='vol1',
                                  display_description='test volume',
                                  size=1,
                                  id='192eb39b-6c2f-420c-bae3-3cfd117f0001',
                                  provider_auth=None,
                                  project_id='project',
                                  volume_type_id=None,
                                  consistencygroup_id=
                                  '192eb39b-6c2f-420c-bae3-3cfd117f0345',
                                  )

    test_snapshot = D()
    test_snapshot.update({'name': 'snapshot1',
                          'size': 1,
                          'volume_size': 1,
                          'id': '192eb39b-6c2f-420c-bae3-3cfd117f0002',
                          'volume_name': 'vol-vol1',
                          'volume_id': '192eb39b-6c2f-420c-bae3-3cfd117f0001',
                          'project_id': 'project',
                          'consistencygroup_id':
                          '192eb39b-6c2f-420c-bae3-3cfd117f0345',
                          })
    test_snapshot.__dict__.update(test_snapshot)
    test_volume2 = {'name': 'vol2',
                    'size': 1,
                    'volume_name': 'vol2',
                    'id': '192eb39b-6c2f-420c-bae3-3cfd117f0004',
                    'provider_auth': None,
                    'project_id': 'project',
                    'display_name': 'vol2',
                    'display_description': 'test volume 2',
                    'volume_type_id': None,
                    'consistencygroup_id':
                    '192eb39b-6c2f-420c-bae3-3cfd117f0345',
                    }
    test_clone = {'name': 'clone1',
                  'size': 1,
                  'volume_name': 'vol3',
                  'id': '192eb39b-6c2f-420c-bae3-3cfd117f0003',
                  'provider_auth': None,
                  'project_id': 'project',
                  'display_name': 'clone1',
                  'display_description': 'volume created from snapshot',
                  'volume_type_id': None,
                  'consistencygroup_id':
                  '192eb39b-6c2f-420c-bae3-3cfd117f0345',
                  }
    unmanaged1 = {'id': 'unmanaged1',
                  'name': 'unmanaged1',
                  'size': 3,
                  }
    group = {'id': '192eb39b-6c2f-420c-bae3-3cfd117f0345',
             'name': 'cg1',
             'status': 'OK',
             }

    cgsnapshot = {
        'id': '192eb39b-6c2f-420c-bae3-3cfd117f9876',
        'consistencygroup_id': group['id'],
        'group_id': None, }

    cgsnapshot_as_group_id = {
        'id': '192eb39b-6c2f-420c-bae3-3cfd117f9876',
        'consistencygroup_id': None,
        'group_id': group['id'], }

    test_volume_attachment = volume_attachment.VolumeAttachment(
        id='2b06255d-f5f0-4520-a953-b029196add6b', volume_id=test_volume.id,
        connector=connector)


class BaseXtremIODriverTestCase(test.TestCase):
    def __init__(self, *args, **kwargs):
        super(BaseXtremIODriverTestCase, self).__init__(*args, **kwargs)
        self.config = mock.Mock(san_login='',
                                san_password='',
                                san_ip='',
                                xtremio_cluster_name='brick1',
                                xtremio_provisioning_factor=20.0,
                                max_over_subscription_ratio=20.0,
                                xtremio_volumes_per_glance_cache=100,
                                driver_ssl_cert_verify=True,
                                driver_ssl_cert_path='/test/path/root_ca.crt',
                                xtremio_array_busy_retry_count=5,
                                xtremio_array_busy_retry_interval=5,
                                xtremio_clean_unused_ig=False)

        def safe_get(key):
            return getattr(self.config, key)
        self.config.safe_get = safe_get

    def setUp(self):
        super(BaseXtremIODriverTestCase, self).setUp()
        clean_xms_data()

        self.driver = xtremio.XtremIOISCSIDriver(configuration=self.config)
        self.driver.client = xtremio.XtremIOClient42(self.config,
                                                     self.config
                                                     .xtremio_cluster_name)
        self.data = CommonData()


@mock.patch('cinder.volume.drivers.dell_emc.xtremio.XtremIOClient.req')
class XtremIODriverISCSITestCase(BaseXtremIODriverTestCase):
    # ##### SetUp Check #####
    def test_check_for_setup_error(self, req):
        req.side_effect = xms_request
        self.driver.check_for_setup_error()
        self.assertEqual(self.driver.client.__class__.__name__,
                         'XtremIOClient42')

    def test_fail_check_for_setup_error(self, req):
        req.side_effect = xms_request
        clusters = xms_data.pop('clusters')
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)
        xms_data['clusters'] = clusters

    def test_check_for_setup_error_ver4(self, req):
        req.side_effect = xms_request
        xms_data['xms'][1]['sw-version'] = '4.0.10-34.hotfix1'
        self.driver.check_for_setup_error()
        self.assertEqual(self.driver.client.__class__.__name__,
                         'XtremIOClient4')

    def test_fail_check_for_array_version(self, req):
        req.side_effect = xms_request
        cluster = xms_data['clusters'][1]
        ver = cluster['sys-sw-version']
        cluster['sys-sw-version'] = '2.0.0-test'
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.check_for_setup_error)
        cluster['sys-sw-version'] = ver

    def test_client4_uses_v2(self, req):
        def base_req(*args, **kwargs):
            self.assertIn('v2', args)
        req.side_effect = base_req
        self.driver.client.req('volumes')

    def test_get_stats(self, req):
        req.side_effect = xms_request
        stats = self.driver.get_volume_stats(True)
        self.assertEqual(self.driver.backend_name,
                         stats['volume_backend_name'])

# ##### Volumes #####
    def test_create_volume_with_cg(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)

    def test_extend_volume(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.driver.extend_volume(self.data.test_volume, 5)

    def test_fail_extend_volume(self, req):
        req.side_effect = xms_request
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.extend_volume, self.data.test_volume, 5)

    def test_delete_volume(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.driver.delete_volume(self.data.test_volume)

    def test_duplicate_volume(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_volume, self.data.test_volume)

# ##### Snapshots #####
    def test_create_snapshot(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.driver.create_snapshot(self.data.test_snapshot)
        self.assertEqual(self.data.test_snapshot['id'],
                         xms_data['volumes'][2]['name'])

    def test_create_delete_snapshot(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.driver.create_snapshot(self.data.test_snapshot)
        self.assertEqual(self.data.test_snapshot['id'],
                         xms_data['volumes'][2]['name'])
        self.driver.delete_snapshot(self.data.test_snapshot)

    def test_failed_rename_snapshot(self, req):
        req.side_effect = xms_failed_rename_snapshot_request
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_snapshot,
                          self.data.test_snapshot)
        self.assertEqual(0, len(xms_data['volumes']))

    def test_volume_from_snapshot(self, req):
        req.side_effect = xms_request
        xms_data['volumes'] = {}
        self.driver.create_volume(self.data.test_volume)
        self.driver.create_snapshot(self.data.test_snapshot)
        self.driver.create_volume_from_snapshot(self.data.test_volume2,
                                                self.data.test_snapshot)

    def test_volume_from_snapshot_and_resize(self, req):
        req.side_effect = xms_request
        xms_data['volumes'] = {}
        self.driver.create_volume(self.data.test_volume)
        clone_volume = self.data.test_clone.copy()
        clone_volume['size'] = 2
        self.driver.create_snapshot(self.data.test_snapshot)
        with mock.patch.object(self.driver,
                               'extend_volume') as extend:
            self.driver.create_volume_from_snapshot(clone_volume,
                                                    self.data.test_snapshot)
            extend.assert_called_once_with(clone_volume, clone_volume['size'])

    def test_volume_from_snapshot_and_resize_fail(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        vol = xms_data['volumes'][1]

        def failed_extend(obj_type='volumes', method='GET', data=None,
                          *args, **kwargs):
            if method == 'GET':
                return {'content': vol}
            elif method == 'POST':
                return {'links': [{'href': 'volume/2'}]}
            elif method == 'PUT':
                if 'name' in data:
                    return
                raise exception.VolumeBackendAPIException('Failed Clone')

        self.driver.create_snapshot(self.data.test_snapshot)
        req.side_effect = failed_extend
        self.driver.db = mock.Mock()
        (self.driver.db.
         image_volume_cache_get_by_volume_id.return_value) = mock.MagicMock()
        clone = self.data.test_clone.copy()
        clone['size'] = 2

        with mock.patch.object(self.driver,
                               'delete_volume') as delete:
            self.assertRaises(exception.VolumeBackendAPIException,
                              self.driver.create_volume_from_snapshot,
                              clone,
                              self.data.test_snapshot)
            self.assertTrue(delete.called)

# ##### Clone Volume #####
    def test_clone_volume(self, req):
        req.side_effect = xms_request
        self.driver.db = mock.Mock()
        (self.driver.db.
         image_volume_cache_get_by_volume_id.return_value) = mock.MagicMock()
        self.driver.create_volume(self.data.test_volume)
        xms_data['volumes'][1]['num-of-dest-snaps'] = 50
        self.driver.create_cloned_volume(self.data.test_clone,
                                         self.data.test_volume)

    def test_clone_volume_exceed_conf_limit(self, req):
        req.side_effect = xms_request
        self.driver.db = mock.Mock()
        (self.driver.db.
         image_volume_cache_get_by_volume_id.return_value) = mock.MagicMock()
        self.driver.create_volume(self.data.test_volume)
        xms_data['volumes'][1]['num-of-dest-snaps'] = 200
        self.assertRaises(exception.CinderException,
                          self.driver.create_cloned_volume,
                          self.data.test_clone,
                          self.data.test_volume)

    @mock.patch.object(xtremio.XtremIOClient4, 'create_snapshot')
    def test_clone_volume_exceed_array_limit(self, create_snap, req):
        create_snap.side_effect = xtremio.XtremIOSnapshotsLimitExceeded()
        req.side_effect = xms_request
        self.driver.db = mock.Mock()
        (self.driver.db.
         image_volume_cache_get_by_volume_id.return_value) = mock.MagicMock()
        self.driver.create_volume(self.data.test_volume)
        xms_data['volumes'][1]['num-of-dest-snaps'] = 50
        self.assertRaises(exception.CinderException,
                          self.driver.create_cloned_volume,
                          self.data.test_clone,
                          self.data.test_volume)

    def test_clone_volume_too_many_snaps(self, req):
        req.side_effect = xms_request
        response = mock.MagicMock()
        response.status_code = 400
        response.json.return_value = {
            "message": "too_many_snapshots_per_vol",
            "error_code": 400
        }
        self.assertRaises(xtremio.XtremIOSnapshotsLimitExceeded,
                          self.driver.client.handle_errors,
                          response, '', '')

    def test_clone_volume_too_many_objs(self, req):
        req.side_effect = xms_request
        response = mock.MagicMock()
        response.status_code = 400
        response.json.return_value = {
            "message": "too_many_objs",
            "error_code": 400
        }
        self.assertRaises(xtremio.XtremIOSnapshotsLimitExceeded,
                          self.driver.client.handle_errors,
                          response, '', '')

    def test_update_migrated_volume(self, req):
        original = self.data.test_volume
        new = self.data.test_volume2
        update = (self.driver.
                  update_migrated_volume({},
                                         original, new, 'available'))
        req.assert_called_once_with('volumes', 'PUT',
                                    {'name': original['id']}, new['id'],
                                    None, 'v2')
        self.assertEqual({'_name_id': None,
                          'provider_location': None}, update)

    def test_update_migrated_volume_failed_rename(self, req):
        req.side_effect = exception.VolumeBackendAPIException(
            data='failed rename')
        original = self.data.test_volume
        new = copy.deepcopy(self.data.test_volume2)
        fake_provider = '__provider'
        new['provider_location'] = fake_provider
        new['_name_id'] = None
        update = (self.driver.
                  update_migrated_volume({},
                                         original, new, 'available'))
        self.assertEqual({'_name_id': new['id'],
                          'provider_location': fake_provider},
                         update)

    def test_clone_volume_and_resize(self, req):
        req.side_effect = xms_request
        self.driver.db = mock.Mock()
        (self.driver.db.
         image_volume_cache_get_by_volume_id.return_value) = mock.MagicMock()
        self.driver.create_volume(self.data.test_volume)
        vol = xms_data['volumes'][1]
        vol['num-of-dest-snaps'] = 0
        clone = self.data.test_clone.copy()
        clone['size'] = 2
        with mock.patch.object(self.driver,
                               'extend_volume') as extend:
            self.driver.create_cloned_volume(clone, self.data.test_volume)
            extend.assert_called_once_with(clone, clone['size'])

    def test_clone_volume_and_resize_fail(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        vol = xms_data['volumes'][1]

        def failed_extend(obj_type='volumes', method='GET', data=None,
                          *args, **kwargs):
            if method == 'GET':
                return {'content': vol}
            elif method == 'POST':
                return {'links': [{'href': 'volume/2'}]}
            elif method == 'PUT':
                if 'name' in data:
                    return
                raise exception.VolumeBackendAPIException('Failed Clone')

        req.side_effect = failed_extend
        self.driver.db = mock.Mock()
        (self.driver.db.
         image_volume_cache_get_by_volume_id.return_value) = mock.MagicMock()
        vol['num-of-dest-snaps'] = 0
        clone = self.data.test_clone.copy()
        clone['size'] = 2
        with mock.patch.object(self.driver,
                               'delete_volume') as delete:
            self.assertRaises(exception.VolumeBackendAPIException,
                              self.driver.create_cloned_volume,
                              clone,
                              self.data.test_volume)
            self.assertTrue(delete.called)

# ##### Connection #####
    def test_no_portals_configured(self, req):
        req.side_effect = xms_request
        portals = xms_data['iscsi-portals'].copy()
        xms_data['iscsi-portals'].clear()
        lunmap = {'lun': 4}
        self.assertRaises(exception.VolumeDriverException,
                          self.driver._get_iscsi_properties, lunmap)
        xms_data['iscsi-portals'] = portals

    def test_initialize_connection(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.driver.create_volume(self.data.test_volume2)
        map_data = self.driver.initialize_connection(self.data.test_volume,
                                                     self.data.connector)
        self.assertEqual(1, map_data['data']['target_lun'])

    def test_initialize_connection_existing_ig(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.driver.create_volume(self.data.test_volume2)
        self.driver.initialize_connection(self.data.test_volume,
                                          self.data.connector)
        i1 = xms_data['initiators'][1]
        i1['ig-id'] = ['', i1['ig-id'], 1]
        i1['chap-authentication-initiator-password'] = 'chap_password1'
        i1['chap-discovery-initiator-password'] = 'chap_password2'
        self.driver.initialize_connection(self.data.test_volume2,
                                          self.data.connector)

    def test_terminate_connection(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.driver.initialize_connection(self.data.test_volume,
                                          self.data.connector)
        i1 = xms_data['initiators'][1]
        i1['ig-id'] = ['', i1['ig-id'], 1]
        self.driver.terminate_connection(self.data.test_volume,
                                         self.data.connector)
        self.assertEqual(1, len(xms_data['initiator-groups']))

    def test_terminate_connection_clean_ig(self, req):
        self.driver.clean_ig = True
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.driver.initialize_connection(self.data.test_volume,
                                          self.data.connector)
        i1 = xms_data['initiators'][1]
        i1['ig-id'] = ['', i1['ig-id'], 1]
        xms_data['initiator-groups'][1]['num-of-vols'] = 0
        # lun mapping list is a list of triplets (IG OID, TG OID, lun number)
        self.driver.terminate_connection(self.data.test_volume,
                                         self.data.connector)
        self.assertEqual(0, len(xms_data['initiator-groups']))

    def test_terminate_connection_fail_on_bad_volume(self, req):
        req.side_effect = xms_request
        self.assertRaises(exception.NotFound,
                          self.driver.terminate_connection,
                          self.data.test_volume,
                          self.data.connector)

    def test_get_ig_indexes_from_initiators_called_once(self, req):
        req.side_effect = xms_request
        volume1 = copy.deepcopy(self.data.test_volume)
        volume1.volume_attachment.objects = [self.data.test_volume_attachment]
        self.driver.create_volume(volume1)
        map_data = self.driver.initialize_connection(self.data.test_volume,
                                                     self.data.connector)
        i1 = xms_data['initiators'][1]
        i1['ig-id'] = ['', i1['ig-id'], 1]
        self.assertEqual(1, map_data['data']['target_lun'])

        with mock.patch.object(self.driver,
                               '_get_ig_indexes_from_initiators') as get_idx:
            get_idx.return_value = [1]
            self.driver.terminate_connection(self.data.test_volume,
                                             self.data.connector)
            get_idx.assert_called_once_with(self.data.connector)

    def test_initialize_connection_after_enabling_chap(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.driver.create_volume(self.data.test_volume2)
        map_data = self.driver.initialize_connection(self.data.test_volume,
                                                     self.data.connector)
        self.assertNotIn('access_mode', map_data['data'])
        c1 = xms_data['clusters'][1]
        c1['chap-authentication-mode'] = 'initiator'
        c1['chap-discovery-mode'] = 'initiator'
        i1 = xms_data['initiators'][1]
        i1['ig-id'] = ['', i1['ig-id'], 1]
        i1['chap-authentication-initiator-password'] = 'chap_password1'
        i1['chap-discovery-initiator-password'] = 'chap_password2'
        map_data = self.driver.initialize_connection(self.data.test_volume2,
                                                     self.data.connector)
        self.assertEqual('chap_password1', map_data['data']['auth_password'])
        self.assertEqual('chap_password2',
                         map_data['data']['discovery_auth_password'])

    def test_initialize_connection_after_disabling_chap(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.driver.create_volume(self.data.test_volume2)
        c1 = xms_data['clusters'][1]
        c1['chap-authentication-mode'] = 'initiator'
        c1['chap-discovery-mode'] = 'initiator'
        self.driver.initialize_connection(self.data.test_volume,
                                          self.data.connector)
        i1 = xms_data['initiators'][1]
        i1['ig-id'] = ['', i1['ig-id'], 1]
        i1['chap-authentication-initiator-password'] = 'chap_password1'
        i1['chap-discovery-initiator-password'] = 'chap_password2'
        i1['chap-authentication-initiator-password'] = None
        i1['chap-discovery-initiator-password'] = None
        self.driver.initialize_connection(self.data.test_volume2,
                                          self.data.connector)

    @mock.patch('oslo_utils.strutils.mask_dict_password')
    def test_initialize_connection_masks_password(self, mask_dict, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.driver.initialize_connection(self.data.test_volume,
                                          self.data.connector)
        self.assertTrue(mask_dict.called)

    def test_add_auth(self, req):
        req.side_effect = xms_request
        data = {}
        self.driver._add_auth(data, True, True)
        self.assertIn('initiator-discovery-user-name', data,
                      'Missing discovery user in data')
        self.assertIn('initiator-discovery-password', data,
                      'Missing discovery password in data')

    def test_initialize_connection_bad_ig(self, req):
        req.side_effect = xms_bad_request
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection,
                          self.data.test_volume,
                          self.data.connector)
        self.driver.delete_volume(self.data.test_volume)

# ##### Manage Volumes #####
    def test_manage_volume(self, req):
        req.side_effect = xms_request
        xms_data['volumes'] = {1: {'name': 'unmanaged1',
                                   'index': 1,
                                   'vol-size': '3',
                                   },
                               }
        ref_vol = {"source-name": "unmanaged1"}
        self.driver.manage_existing(self.data.test_volume, ref_vol)

    def test_failed_manage_volume(self, req):
        req.side_effect = xms_request
        xms_data['volumes'] = {1: {'name': 'unmanaged1',
                                   'index': 1,
                                   'vol-size': '3',
                                   },
                               }
        invalid_ref = {"source-name": "invalid"}
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing,
                          self.data.test_volume, invalid_ref)

    def test_get_manage_volume_size(self, req):
        req.side_effect = xms_request
        xms_data['volumes'] = {1: {'name': 'unmanaged1',
                                   'index': 1,
                                   'vol-size': '1000000',
                                   },
                               }
        ref_vol = {"source-name": "unmanaged1"}
        size = self.driver.manage_existing_get_size(self.data.test_volume,
                                                    ref_vol)
        self.assertEqual(1, size)

    def test_manage_volume_size_invalid_input(self, req):
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_get_size,
                          self.data.test_volume, {})

    def test_failed_manage_volume_size(self, req):
        req.side_effect = xms_request
        xms_data['volumes'] = {1: {'name': 'unmanaged1',
                                   'index': 1,
                                   'vol-size': '3',
                                   },
                               }
        invalid_ref = {"source-name": "invalid"}
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_get_size,
                          self.data.test_volume, invalid_ref)

    def test_unmanage_volume(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.driver.unmanage(self.data.test_volume)

    def test_failed_unmanage_volume(self, req):
        req.side_effect = xms_request
        self.assertRaises(exception.VolumeNotFound, self.driver.unmanage,
                          self.data.test_volume2)

    def test_manage_snapshot(self, req):
        req.side_effect = xms_request
        vol_uid = self.data.test_snapshot.volume_id
        xms_data['volumes'] = {1: {'name': vol_uid,
                                   'index': 1,
                                   'vol-size': '3',
                                   },
                               2: {'name': 'unmanaged',
                                   'index': 2,
                                   'ancestor-vol-id': ['', vol_uid, 1],
                                   'vol-size': '3'}
                               }
        ref_vol = {"source-name": "unmanaged"}
        self.driver.manage_existing_snapshot(self.data.test_snapshot, ref_vol)

    def test_get_manage_snapshot_size(self, req):
        req.side_effect = xms_request
        vol_uid = self.data.test_snapshot.volume_id
        xms_data['volumes'] = {1: {'name': vol_uid,
                                   'index': 1,
                                   'vol-size': '3',
                                   },
                               2: {'name': 'unmanaged',
                                   'index': 2,
                                   'ancestor-vol-id': ['', vol_uid, 1],
                                   'vol-size': '3'}
                               }
        ref_vol = {"source-name": "unmanaged"}
        self.driver.manage_existing_snapshot_get_size(self.data.test_snapshot,
                                                      ref_vol)

    def test_manage_snapshot_invalid_snapshot(self, req):
        req.side_effect = xms_request
        xms_data['volumes'] = {1: {'name': 'unmanaged1',
                                   'index': 1,
                                   'vol-size': '3',
                                   'ancestor-vol-id': []}
                               }
        ref_vol = {"source-name": "unmanaged1"}
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_snapshot,
                          self.data.test_snapshot, ref_vol)

    def test_unmanage_snapshot(self, req):
        req.side_effect = xms_request
        vol_uid = self.data.test_snapshot.volume_id
        xms_data['volumes'] = {1: {'name': vol_uid,
                                   'index': 1,
                                   'vol-size': '3',
                                   },
                               2: {'name': 'unmanaged',
                                   'index': 2,
                                   'ancestor-vol-id': ['', vol_uid, 1],
                                   'vol-size': '3'}
                               }
        ref_vol = {"source-name": "unmanaged"}
        self.driver.manage_existing_snapshot(self.data.test_snapshot, ref_vol)
        self.driver.unmanage_snapshot(self.data.test_snapshot)

# ##### Consistancy Groups #####
    @mock.patch('cinder.objects.snapshot.SnapshotList.get_all_for_cgsnapshot')
    def test_cg_create(self, get_all_for_cgsnapshot, req):
        req.side_effect = xms_request
        d = self.data
        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']
        get_all_for_cgsnapshot.return_value = [snapshot_obj]

        self.driver.create_consistencygroup(d.context, d.group)
        self.assertEqual(1, len(xms_data['consistency-groups']))

    @mock.patch('cinder.objects.snapshot.SnapshotList.get_all_for_cgsnapshot')
    def test_cg_update(self, get_all_for_cgsnapshot, req):
        req.side_effect = xms_request
        d = self.data
        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']
        get_all_for_cgsnapshot.return_value = [snapshot_obj]

        self.driver.create_consistencygroup(d.context, d.group)
        self.driver.update_consistencygroup(d.context, d.group,
                                            add_volumes=[d.test_volume,
                                                         d.test_volume2])
        self.assertEqual(2, len(xms_data['consistency-group-volumes']))
        self.driver.update_consistencygroup(d.context, d.group,
                                            remove_volumes=[d.test_volume2])
        self.assertEqual(1, len(xms_data['consistency-group-volumes']))

    @mock.patch('cinder.objects.snapshot.SnapshotList.get_all_for_cgsnapshot')
    def test_create_cg(self, get_all_for_cgsnapshot, req):
        req.side_effect = xms_request
        d = self.data
        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']
        get_all_for_cgsnapshot.return_value = [snapshot_obj]
        self.driver.create_consistencygroup(d.context, d.group)
        self.driver.update_consistencygroup(d.context, d.group,
                                            add_volumes=[d.test_volume,
                                                         d.test_volume2])
        self.driver.db = mock.Mock()
        (self.driver.db.
         volume_get_all_by_group.return_value) = [mock.MagicMock()]
        res = self.driver.create_cgsnapshot(d.context, d.cgsnapshot,
                                            [snapshot_obj])
        self.assertEqual((None, None), res)

    @mock.patch('cinder.objects.snapshot.SnapshotList.get_all_for_cgsnapshot')
    def test_cg_delete(self, get_all_for_cgsnapshot, req):
        req.side_effect = xms_request
        d = self.data
        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']
        get_all_for_cgsnapshot.return_value = [snapshot_obj]
        self.driver.create_consistencygroup(d.context, d.group)
        self.driver.update_consistencygroup(d.context, d.group,
                                            add_volumes=[d.test_volume,
                                                         d.test_volume2])
        self.driver.db = mock.Mock()
        self.driver.create_cgsnapshot(d.context, d.cgsnapshot, [snapshot_obj])
        self.driver.delete_consistencygroup(d.context, d.group, [])

    def test_cg_delete_with_volume(self, req):
        req.side_effect = xms_request
        d = self.data
        self.driver.create_consistencygroup(d.context, d.group)
        self.driver.create_volume(d.test_volume)
        self.driver.update_consistencygroup(d.context, d.group,
                                            add_volumes=[d.test_volume])
        self.driver.db = mock.Mock()

        results, volumes = \
            self.driver.delete_consistencygroup(d.context,
                                                d.group,
                                                [d.test_volume])

        self.assertTrue(all(volume['status'] == 'deleted' for volume in
                            volumes))

    @mock.patch('cinder.objects.snapshot.SnapshotList.get_all_for_cgsnapshot')
    def test_cg_snapshot(self, get_all_for_cgsnapshot, req):
        req.side_effect = xms_request
        d = self.data
        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']
        get_all_for_cgsnapshot.return_value = [snapshot_obj]
        self.driver.create_consistencygroup(d.context, d.group)
        self.driver.update_consistencygroup(d.context, d.group,
                                            add_volumes=[d.test_volume,
                                                         d.test_volume2])
        snapset_name = self.driver._get_cgsnap_name(d.cgsnapshot)
        self.assertEqual(snapset_name,
                         '192eb39b6c2f420cbae33cfd117f0345192eb39b6c2f420cbae'
                         '33cfd117f9876')
        snapset1 = {'ancestor-vol-id': ['', d.test_volume['id'], 2],
                    'consistencygroup_id': d.group['id'],
                    'name': snapset_name,
                    'index': 1}
        xms_data['snapshot-sets'] = {snapset_name: snapset1, 1: snapset1}
        res = self.driver.delete_cgsnapshot(d.context, d.cgsnapshot,
                                            [snapshot_obj])
        self.assertEqual((None, None), res)

    def test_delete_cgsnapshot(self, req):
        d = self.data
        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']
        self.driver.delete_cgsnapshot(d.context, d.cgsnapshot,
                                      [snapshot_obj])
        req.assert_called_once_with('snapshot-sets', 'DELETE', None,
                                    '192eb39b6c2f420cbae33cfd117f0345192eb39'
                                    'b6c2f420cbae33cfd117f9876', None, 'v2')

    @mock.patch('cinder.objects.snapshot.SnapshotList.get_all_for_cgsnapshot')
    def test_cg_from_src_snapshot(self, get_all_for_cgsnapshot, req):
        req.side_effect = xms_request
        d = self.data

        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']
        snapshot_obj.volume_id = d.test_volume['id']
        get_all_for_cgsnapshot.return_value = [snapshot_obj]

        self.driver.create_consistencygroup(d.context, d.group)
        self.driver.create_volume(d.test_volume)
        self.driver.create_cgsnapshot(d.context, d.cgsnapshot, [])
        xms_data['volumes'][2]['ancestor-vol-id'] = (xms_data['volumes'][1]
                                                     ['vol-id'])
        snapset_name = self.driver._get_cgsnap_name(d.cgsnapshot)

        snapset1 = {'vol-list': [xms_data['volumes'][2]['vol-id']],
                    'name': snapset_name,
                    'index': 1}
        xms_data['snapshot-sets'] = {snapset_name: snapset1, 1: snapset1}
        cg_obj = fake_cg.fake_consistencyobject_obj(d.context)
        new_vol1 = fake_volume_obj(d.context)
        snapshot1 = (fake_snapshot
                     .fake_snapshot_obj
                     (d.context, volume_id=d.test_volume['id']))
        res = self.driver.create_consistencygroup_from_src(d.context, cg_obj,
                                                           [new_vol1],
                                                           d.cgsnapshot,
                                                           [snapshot1])
        self.assertEqual((None, None), res)

    @mock.patch('cinder.objects.snapshot.SnapshotList.get_all_for_cgsnapshot')
    def test_cg_from_src_cg(self, get_all_for_cgsnapshot, req):
        req.side_effect = xms_request
        d = self.data

        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']
        snapshot_obj.volume_id = d.test_volume['id']
        get_all_for_cgsnapshot.return_value = [snapshot_obj]

        self.driver.create_consistencygroup(d.context, d.group)
        self.driver.create_volume(d.test_volume)
        self.driver.create_cgsnapshot(d.context, d.cgsnapshot, [])
        xms_data['volumes'][2]['ancestor-vol-id'] = (xms_data['volumes'][1]
                                                     ['vol-id'])
        snapset_name = self.driver._get_cgsnap_name(d.cgsnapshot)

        snapset1 = {'vol-list': [xms_data['volumes'][2]['vol-id']],
                    'name': snapset_name,
                    'index': 1}
        xms_data['snapshot-sets'] = {snapset_name: snapset1, 1: snapset1}
        cg_obj = fake_cg.fake_consistencyobject_obj(d.context)
        new_vol1 = fake_volume_obj(d.context)
        new_cg_obj = fake_cg.fake_consistencyobject_obj(
            d.context, id=fake.CONSISTENCY_GROUP2_ID)
        snapset2_name = new_cg_obj.id
        new_vol1.id = '192eb39b-6c2f-420c-bae3-3cfd117f0001'
        new_vol2 = fake_volume_obj(d.context)
        snapset2 = {'vol-list': [xms_data['volumes'][2]['vol-id']],
                    'name': snapset2_name,
                    'index': 1}
        xms_data['snapshot-sets'].update({5: snapset2,
                                          snapset2_name: snapset2})
        self.driver.create_consistencygroup_from_src(d.context, new_cg_obj,
                                                     [new_vol2],
                                                     None, None,
                                                     cg_obj, [new_vol1])

    @mock.patch('cinder.objects.snapshot.SnapshotList.get_all_for_cgsnapshot')
    def test_invalid_cg_from_src_input(self, get_all_for_cgsnapshot, req):
        req.side_effect = xms_request
        d = self.data

        self.assertRaises(exception.InvalidInput,
                          self.driver.create_consistencygroup_from_src,
                          d.context, d.group, [], None, None, None, None)

#   #### Groups ####
    def test_group_create(self, req):
        """Test group create."""

        req.side_effect = xms_request
        d = self.data

        self.driver.create_group(d.context, d.group)
        self.assertEqual(1, len(xms_data['consistency-groups']))

    def test_group_update(self, req):
        """Test group update."""

        req.side_effect = xms_request
        d = self.data

        self.driver.create_consistencygroup(d.context, d.group)
        self.driver.update_consistencygroup(d.context, d.group,
                                            add_volumes=[d.test_volume,
                                                         d.test_volume2])
        self.assertEqual(2, len(xms_data['consistency-group-volumes']))
        self.driver.update_group(d.context, d.group,
                                 remove_volumes=[d.test_volume2])
        self.assertEqual(1, len(xms_data['consistency-group-volumes']))

    def test_create_group_snapshot(self, req):
        """Test create group snapshot."""

        req.side_effect = xms_request
        d = self.data
        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']

        self.driver.create_group(d.context, d.group)
        self.driver.update_group(d.context, d.group,
                                 add_volumes=[d.test_volume,
                                              d.test_volume2])

        res = self.driver.create_group_snapshot(d.context, d.cgsnapshot,
                                                [snapshot_obj])
        self.assertEqual((None, None), res)

    def test_group_delete(self, req):
        """"Test delete group."""
        req.side_effect = xms_request
        d = self.data
        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']
        self.driver.create_group(d.context, d.group)
        self.driver.update_group(d.context, d.group,
                                 add_volumes=[d.test_volume,
                                              d.test_volume2])
        self.driver.db = mock.Mock()
        (self.driver.db.
         volume_get_all_by_group.return_value) = [mock.MagicMock()]
        self.driver.create_group_snapshot(d.context, d.cgsnapshot,
                                          [snapshot_obj])
        self.driver.delete_group(d.context, d.group, [])

    def test_group_delete_with_volume(self, req):
        req.side_effect = xms_request
        d = self.data
        self.driver.create_consistencygroup(d.context, d.group)
        self.driver.create_volume(d.test_volume)
        self.driver.update_consistencygroup(d.context, d.group,
                                            add_volumes=[d.test_volume])
        self.driver.db = mock.Mock()

        results, volumes = \
            self.driver.delete_group(d.context, d.group, [d.test_volume])

        self.assertTrue(all(volume['status'] == 'deleted' for volume in
                            volumes))

    def test_group_snapshot(self, req):
        """test group snapshot."""
        req.side_effect = xms_request
        d = self.data
        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']
        self.driver.create_group(d.context, d.group)
        self.driver.update_group(d.context, d.group,
                                 add_volumes=[d.test_volume,
                                              d.test_volume2])

        snapset_name = self.driver._get_cgsnap_name(d.cgsnapshot)
        self.assertEqual(snapset_name,
                         '192eb39b6c2f420cbae33cfd117f0345192eb39b6c2f420cbae'
                         '33cfd117f9876')
        snapset1 = {'ancestor-vol-id': ['', d.test_volume['id'], 2],
                    'consistencygroup_id': d.group['id'],
                    'name': snapset_name,
                    'index': 1}
        xms_data['snapshot-sets'] = {snapset_name: snapset1, 1: snapset1}
        res = self.driver.delete_group_snapshot(d.context, d.cgsnapshot,
                                                [snapshot_obj])
        self.assertEqual((None, None), res)

    def test_group_snapshot_with_generic_group(self, req):
        """test group snapshot shot with generic group ."""
        req.side_effect = xms_request
        d = self.data
        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']
        self.driver.create_group(d.context, d.group)
        self.driver.update_group(d.context, d.group,
                                 add_volumes=[d.test_volume,
                                              d.test_volume2])

        snapset_name = self.driver._get_cgsnap_name(d.cgsnapshot_as_group_id)
        self.assertEqual(snapset_name,
                         '192eb39b6c2f420cbae33cfd117f0345192eb39b6c2f420cbae'
                         '33cfd117f9876')
        snapset1 = {'ancestor-vol-id': ['', d.test_volume['id'], 2],
                    'consistencygroup_id': d.group['id'],
                    'name': snapset_name,
                    'index': 1}
        xms_data['snapshot-sets'] = {snapset_name: snapset1, 1: snapset1}
        res = self.driver.delete_group_snapshot(d.context, d.cgsnapshot,
                                                [snapshot_obj])
        self.assertEqual((None, None), res)

    def test_delete_group_snapshot(self, req):
        """test delete group snapshot."""
        d = self.data
        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']
        self.driver.delete_group_snapshot(d.context, d.cgsnapshot,
                                          [snapshot_obj])
        req.assert_called_once_with('snapshot-sets', 'DELETE', None,
                                    '192eb39b6c2f420cbae33cfd117f0345192eb39'
                                    'b6c2f420cbae33cfd117f9876', None, 'v2')

    def test_delete_group_snapshot_with_generic_group(self, req):
        """test delete group snapshot."""
        d = self.data
        snapshot_obj = fake_snapshot.fake_snapshot_obj(d.context)
        snapshot_obj.consistencygroup_id = d.group['id']
        self.driver.delete_group_snapshot(d.context, d.cgsnapshot_as_group_id,
                                          [snapshot_obj])
        req.assert_called_once_with('snapshot-sets', 'DELETE', None,
                                    '192eb39b6c2f420cbae33cfd117f0345192eb39'
                                    'b6c2f420cbae33cfd117f9876', None, 'v2')

    def test_group_from_src_snapshot(self, req):
        """test group from source snapshot."""
        req.side_effect = xms_request
        d = self.data

        self.driver.create_group(d.context, d.group)
        self.driver.create_volume(d.test_volume)
        self.driver.create_group_snapshot(d.context, d.cgsnapshot, [])
        xms_data['volumes'][2]['ancestor-vol-id'] = (xms_data['volumes'][1]
                                                     ['vol-id'])
        snapset_name = self.driver._get_cgsnap_name(d.cgsnapshot)

        snapset1 = {'vol-list': [xms_data['volumes'][2]['vol-id']],
                    'name': snapset_name,
                    'index': 1}
        xms_data['snapshot-sets'] = {snapset_name: snapset1, 1: snapset1}
        cg_obj = fake_cg.fake_consistencyobject_obj(d.context)
        new_vol1 = fake_volume_obj(d.context)
        snapshot1 = (fake_snapshot
                     .fake_snapshot_obj
                     (d.context, volume_id=d.test_volume['id']))
        res = self.driver.create_group_from_src(d.context, cg_obj,
                                                [new_vol1],
                                                d.cgsnapshot,
                                                [snapshot1])
        self.assertEqual((None, None), res)

    def test_group_from_src_group(self, req):
        """test group from source group."""
        req.side_effect = xms_request
        d = self.data

        self.driver.create_group(d.context, d.group)
        self.driver.create_volume(d.test_volume)
        self.driver.create_group_snapshot(d.context, d.cgsnapshot, [])
        xms_data['volumes'][2]['ancestor-vol-id'] = (xms_data['volumes'][1]
                                                     ['vol-id'])
        snapset_name = self.driver._get_cgsnap_name(d.cgsnapshot)

        snapset1 = {'vol-list': [xms_data['volumes'][2]['vol-id']],
                    'name': snapset_name,
                    'index': 1}
        xms_data['snapshot-sets'] = {snapset_name: snapset1, 1: snapset1}
        cg_obj = fake_cg.fake_consistencyobject_obj(d.context)
        new_vol1 = fake_volume_obj(d.context)
        new_cg_obj = fake_cg.fake_consistencyobject_obj(
            d.context, id=fake.CONSISTENCY_GROUP2_ID)
        snapset2_name = new_cg_obj.id
        new_vol1.id = '192eb39b-6c2f-420c-bae3-3cfd117f0001'
        new_vol2 = fake_volume_obj(d.context)
        snapset2 = {'vol-list': [xms_data['volumes'][2]['vol-id']],
                    'name': snapset2_name,
                    'index': 1}
        xms_data['snapshot-sets'].update({5: snapset2,
                                          snapset2_name: snapset2})
        self.driver.create_group_from_src(d.context, new_cg_obj,
                                          [new_vol2],
                                          None, None,
                                          cg_obj, [new_vol1])

    def test_invalid_group_from_src_input(self, req):
        """test invalid group from source."""
        req.side_effect = xms_request
        d = self.data

        self.assertRaises(exception.InvalidInput,
                          self.driver.create_group_from_src,
                          d.context, d.group, [], None, None, None, None)

    def test_get_password(self, _req):
        p = self.driver._get_password()
        self.assertEqual(len(p), 12)
        self.assertIsNotNone(re.match(r'[A-Z0-9]{12}', p), p)


@mock.patch('requests.request')
class XtremIODriverTestCase(BaseXtremIODriverTestCase):
    # ##### XMS Client #####
    @mock.patch.object(time, 'sleep', mock.Mock(return_value=0))
    def test_retry_request(self, req):
        busy_response = mock.MagicMock()
        busy_response.status_code = 400
        busy_response.json.return_value = {
            "message": "system_is_busy",
            "error_code": 400
        }
        good_response = mock.MagicMock()
        good_response.status_code = 200

        XtremIODriverTestCase.req_count = 0

        def busy_request(*args, **kwargs):
            if XtremIODriverTestCase.req_count < 1:
                XtremIODriverTestCase.req_count += 1
                return busy_response
            return good_response

        req.side_effect = busy_request
        self.driver.create_volume(self.data.test_volume)

    def test_verify_cert(self, req):
        good_response = mock.MagicMock()
        good_response.status_code = 200

        def request_verify_cert(*args, **kwargs):
            self.assertEqual(kwargs['verify'], '/test/path/root_ca.crt')
            return good_response

        req.side_effect = request_verify_cert
        self.driver.client.req('volumes')


@mock.patch('cinder.volume.drivers.dell_emc.xtremio.XtremIOClient.req')
class XtremIODriverFCTestCase(BaseXtremIODriverTestCase):
    def setUp(self):
        super(XtremIODriverFCTestCase, self).setUp()
        self.driver = xtremio.XtremIOFCDriver(
            configuration=self.config)

# ##### Connection FC#####
    def test_initialize_connection(self, req):
        req.side_effect = xms_request

        self.driver.create_volume(self.data.test_volume)
        map_data = self.driver.initialize_connection(self.data.test_volume,
                                                     self.data.connector)
        self.assertEqual(1, map_data['data']['target_lun'])

    def test_terminate_connection(self, req):
        req.side_effect = xms_request

        self.driver.create_volume(self.data.test_volume)
        self.driver.initialize_connection(self.data.test_volume,
                                          self.data.connector)
        for i1 in xms_data['initiators'].values():
            i1['ig-id'] = ['', i1['ig-id'], 1]
        self.driver.terminate_connection(self.data.test_volume,
                                         self.data.connector)

    def test_force_terminate_connection(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)
        self.driver.initialize_connection(self.data.test_volume,
                                          self.data.connector)
        vol1 = xms_data['volumes'][1]
        # lun mapping list is a list of triplets (IG OID, TG OID, lun number)
        vol1['lun-mapping-list'] = [[['a91e8c81c2d14ae4865187ce4f866f8a',
                                      'iqn.1993-08.org.debian:01:222',
                                      1],
                                     ['', 'Default', 1],
                                    1]]
        self.driver.terminate_connection(self.data.test_volume, None)

    def test_initialize_existing_ig_connection(self, req):
        req.side_effect = xms_request
        self.driver.create_volume(self.data.test_volume)

        pre_existing = 'pre_existing_host'
        self.driver._create_ig(pre_existing)
        wwpns = self.driver._get_initiator_names(self.data.connector)
        for wwpn in wwpns:
            data = {'initiator-name': wwpn, 'ig-id': pre_existing,
                    'port-address': wwpn}
            self.driver.client.req('initiators', 'POST', data)

        def get_fake_initiator(wwpn):
            return {'port-address': wwpn, 'ig-id': ['', pre_existing, 1]}
        with mock.patch.object(self.driver.client, 'get_initiator',
                               side_effect=get_fake_initiator):
            map_data = self.driver.initialize_connection(self.data.test_volume,
                                                         self.data.connector)
        self.assertEqual(1, map_data['data']['target_lun'])
        self.assertEqual(1, len(xms_data['initiator-groups']))

    def test_get_initiator_igs_ver4(self, req):
        req.side_effect = xms_request
        wwpn1 = '11:22:33:44:55:66:77:88'
        wwpn2 = '11:22:33:44:55:66:77:89'
        port_addresses = [wwpn1, wwpn2]
        ig_id = ['', 'my_ig', 1]
        self.driver.client = xtremio.XtremIOClient4(self.config,
                                                    self.config
                                                    .xtremio_cluster_name)

        def get_fake_initiator(wwpn):
            return {'port-address': wwpn, 'ig-id': ig_id}
        with mock.patch.object(self.driver.client, 'get_initiator',
                               side_effect=get_fake_initiator):
            self.driver.client.get_initiators_igs(port_addresses)

    def test_get_free_lun(self, req):
        def lm_response(*args, **kwargs):
            return {'lun-maps': [{'lun': 1}]}
        req.side_effect = lm_response

        ig_names = ['test1', 'test2']
        self.driver._get_free_lun(ig_names)

    def test_race_on_terminate_connection(self, req):
        """Test for race conditions on num_of_mapped_volumes.

        This test confirms that num_of_mapped_volumes won't break even if we
        receive a NotFound exception when retrieving info on a specific
        mapping, as that specific mapping could have been deleted between
        the request to get the list of exiting mappings and the request to get
        the info on one of them.
        """
        req.side_effect = xms_request
        self.driver.client = xtremio.XtremIOClient3(
            self.config, self.config.xtremio_cluster_name)
        # We'll wrap num_of_mapped_volumes, we'll store here original method
        original_method = self.driver.client.num_of_mapped_volumes

        def fake_num_of_mapped_volumes(*args, **kwargs):
            # Add a nonexistent mapping
            mappings = [{'href': 'volumes/1'}, {'href': 'volumes/12'}]

            # Side effects will be: 1st call returns the list, then we return
            # data for existing mappings, and on the nonexistent one we added
            # we return NotFound
            side_effect = [{'lun-maps': mappings},
                           {'content': xms_data['lun-maps'][1]},
                           exception.NotFound]

            with mock.patch.object(self.driver.client, 'req',
                                   side_effect=side_effect):
                return original_method(*args, **kwargs)

        self.driver.create_volume(self.data.test_volume)
        map_data = self.driver.initialize_connection(self.data.test_volume,
                                                     self.data.connector)
        self.assertEqual(1, map_data['data']['target_lun'])
        with mock.patch.object(self.driver.client, 'num_of_mapped_volumes',
                               side_effect=fake_num_of_mapped_volumes):
            self.driver.terminate_connection(self.data.test_volume,
                                             self.data.connector)
        self.driver.delete_volume(self.data.test_volume)



# Copyright (c) 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ast
import re

from hacking import core
import six

"""
Guidelines for writing new hacking checks

 - Use only for Cinder specific tests. OpenStack general tests
   should be submitted to the common 'hacking' module.
 - Pick numbers in the range N3xx. Find the current test with
   the highest allocated number and then pick the next value.
 - Keep the test method code in the source file ordered based
   on the N3xx value.
 - List the new rule in the top level HACKING.rst file
 - Add test cases for each new rule to
   cinder/tests/unit/test_hacking.py

"""

# NOTE(thangp): Ignore N323 pep8 error caused by importing cinder objects
UNDERSCORE_IMPORT_FILES = ['cinder/objects/__init__.py',
                           'cinder/objects/manageableresources.py']

mutable_default_args = re.compile(r"^\s*def .+\((.+=\{\}|.+=\[\])")
translated_log = re.compile(
    r"(.)*LOG\.(audit|debug|error|info|warn|warning|critical|exception)"
    r"\(\s*_\(\s*('|\")")
string_translation = re.compile(r"(.)*_\(\s*('|\")")
underscore_import_check = re.compile(r"(.)*i18n\s+import(.)* _$")
underscore_import_check_multi = re.compile(r"(.)*i18n\s+import(.)* _, (.)*")
# We need this for cases where they have created their own _ function.
custom_underscore_check = re.compile(r"(.)*_\s*=\s*(.)*")
no_print_statements = re.compile(r"\s*print\s*\(.+\).*")
dict_constructor_with_list_copy_re = re.compile(r".*\bdict\((\[)?(\(|\[)")

logging_instance = re.compile(
    r"(.)*LOG\.(warning|info|debug|error|exception)\(")

assert_True = re.compile(
    r".*assertEqual\(True, .*\)")


class BaseASTChecker(ast.NodeVisitor):
    """Provides a simple framework for writing AST-based checks.

    Subclasses should implement visit_* methods like any other AST visitor
    implementation. When they detect an error for a particular node the
    method should call ``self.add_error(offending_node)``. Details about
    where in the code the error occurred will be pulled from the node
    object.

    Subclasses should also provide a class variable named CHECK_DESC to
    be used for the human readable error message.

    """

    def __init__(self, tree, filename):
        """This object is created automatically by pep8.

        :param tree: an AST tree
        :param filename: name of the file being analyzed
                         (ignored by our checks)
        """
        self._tree = tree
        self._errors = []

    def run(self):
        """Called automatically by pep8."""
        self.visit(self._tree)
        return self._errors

    def add_error(self, node, message=None):
        """Add an error caused by a node to the list of errors for pep8."""

        # Need to disable pylint check here as it doesn't catch CHECK_DESC
        # being defined in the subclasses.
        message = message or self.CHECK_DESC  # pylint: disable=E1101
        error = (node.lineno, node.col_offset, message, self.__class__)
        self._errors.append(error)

    def _check_call_names(self, call_node, names):
        if isinstance(call_node, ast.Call):
            if isinstance(call_node.func, ast.Name):
                if call_node.func.id in names:
                    return True
        return False


@core.flake8ext
def no_translate_logs(logical_line, filename):
    """Check for 'LOG.*(_('

    Starting with the Pike series, OpenStack no longer supports log
    translation. We shouldn't translate logs.

    - This check assumes that 'LOG' is a logger.
    - Use filename so we can start enforcing this in specific folders
      instead of needing to do so all at once.

    C312
    """
    if translated_log.match(logical_line):
        yield(0, "C312: Log messages should not be translated!")


@core.flake8ext
def no_mutable_default_args(logical_line):
    msg = "N322: Method's default argument shouldn't be mutable!"
    if mutable_default_args.match(logical_line):
        yield (0, msg)


@core.flake8ext
def check_explicit_underscore_import(logical_line, filename):
    """Check for explicit import of the _ function

    We need to ensure that any files that are using the _() function
    to translate messages are explicitly importing the _ function.  We
    can't trust unit test to catch whether the import has been
    added so we need to check for it here.
    """

    # Build a list of the files that have _ imported.  No further
    # checking needed once it is found.
    for file in UNDERSCORE_IMPORT_FILES:
        if file in filename:
            return
    if (underscore_import_check.match(logical_line) or
            underscore_import_check_multi.match(logical_line) or
            custom_underscore_check.match(logical_line)):
        UNDERSCORE_IMPORT_FILES.append(filename)
    elif string_translation.match(logical_line):
        yield(0, "N323: Found use of _() without explicit import of _ !")


class CheckLoggingFormatArgs(BaseASTChecker):
    """Check for improper use of logging format arguments.

    LOG.debug("Volume %s caught fire and is at %d degrees C and climbing.",
              ('volume1', 500))

    The format arguments should not be a tuple as it is easy to miss.

    """

    name = 'check_logging_format_args'
    version = '1.0'

    CHECK_DESC = 'C310 Log method arguments should not be a tuple.'
    LOG_METHODS = [
        'debug', 'info',
        'warn', 'warning',
        'error', 'exception',
        'critical', 'fatal',
        'trace', 'log'
    ]

    def _find_name(self, node):
        """Return the fully qualified name or a Name or Attribute."""
        if isinstance(node, ast.Name):
            return node.id
        elif (isinstance(node, ast.Attribute)
                and isinstance(node.value, (ast.Name, ast.Attribute))):
            method_name = node.attr
            obj_name = self._find_name(node.value)
            if obj_name is None:
                return None
            return obj_name + '.' + method_name
        elif isinstance(node, six.string_types):
            return node
        else:  # could be Subscript, Call or many more
            return None

    def visit_Call(self, node):
        """Look for the 'LOG.*' calls."""
        # extract the obj_name and method_name
        if isinstance(node.func, ast.Attribute):
            obj_name = self._find_name(node.func.value)
            if isinstance(node.func.value, ast.Name):
                method_name = node.func.attr
            elif isinstance(node.func.value, ast.Attribute):
                obj_name = self._find_name(node.func.value)
                method_name = node.func.attr
            else:  # could be Subscript, Call or many more
                return super(CheckLoggingFormatArgs, self).generic_visit(node)

            # obj must be a logger instance and method must be a log helper
            if (obj_name != 'LOG'
                    or method_name not in self.LOG_METHODS):
                return super(CheckLoggingFormatArgs, self).generic_visit(node)

            # the call must have arguments
            if not len(node.args):
                return super(CheckLoggingFormatArgs, self).generic_visit(node)

            # any argument should not be a tuple
            for arg in node.args:
                if isinstance(arg, ast.Tuple):
                    self.add_error(arg)

        return super(CheckLoggingFormatArgs, self).generic_visit(node)


class CheckOptRegistrationArgs(BaseASTChecker):
    """Verifying the registration of options are well formed

    This class creates a check for single opt or list/tuple of
    opts when register_opt() or register_opts() are being called.
    """

    name = 'check_opt_registrationg_args'
    version = '1.0'

    CHECK_DESC = ('C311: Arguments being passed to register_opt/register_opts '
                  'must be a single option or list/tuple of options '
                  'respectively. Options must also end with _opt or _opts '
                  'respectively.')

    singular_method = 'register_opt'
    plural_method = 'register_opts'

    register_methods = [
        singular_method,
        plural_method,
    ]

    def _find_name(self, node):
        """Return the fully qualified name or a Name or Attribute."""
        if isinstance(node, ast.Name):
            return node.id
        elif (isinstance(node, ast.Attribute)
                and isinstance(node.value, (ast.Name, ast.Attribute))):
            method_name = node.attr
            obj_name = self._find_name(node.value)
            if obj_name is None:
                return None
            return obj_name + '.' + method_name
        elif isinstance(node, six.string_types):
            return node
        else:  # could be Subscript, Call or many more
            return None

    def _is_list_or_tuple(self, obj):
        return isinstance(obj, (ast.List, ast.Tuple))

    def visit_Call(self, node):
        """Look for the register_opt/register_opts calls."""
        # extract the obj_name and method_name
        if isinstance(node.func, ast.Attribute):
            if not isinstance(node.func.value, ast.Name):
                return (super(CheckOptRegistrationArgs,
                              self).generic_visit(node))

            method_name = node.func.attr

            # obj must be instance of register_opt() or register_opts()
            if method_name not in self.register_methods:
                return (super(CheckOptRegistrationArgs,
                              self).generic_visit(node))

            if len(node.args) > 0:
                argument_name = self._find_name(node.args[0])
                if argument_name:
                    if (method_name == self.singular_method and
                            not argument_name.lower().endswith('opt')):
                        self.add_error(node.args[0])
                    elif (method_name == self.plural_method and
                            not argument_name.lower().endswith('opts')):
                        self.add_error(node.args[0])
                else:
                    # This covers instances of register_opt()/register_opts()
                    # that are registering the objects directly and not
                    # passing in a variable referencing the options being
                    # registered.
                    if (method_name == self.singular_method and
                            self._is_list_or_tuple(node.args[0])):
                        self.add_error(node.args[0])
                    elif (method_name == self.plural_method and not
                            self._is_list_or_tuple(node.args[0])):
                        self.add_error(node.args[0])

        return super(CheckOptRegistrationArgs, self).generic_visit(node)


@core.flake8ext
def check_datetime_now(logical_line, noqa):
    if noqa:
        return

    msg = ("C301: Found datetime.now(). "
           "Please use timeutils.utcnow() from oslo_utils.")
    if 'datetime.now' in logical_line:
        yield(0, msg)


@core.flake8ext
def check_no_print_statements(logical_line, filename, noqa):
    # CLI and utils programs do need to use 'print()' so
    # we shouldn't check those files.
    if noqa:
        return

    if "cinder/cmd" in filename or "tools/" in filename:
        return

    if re.match(no_print_statements, logical_line):
        msg = ("C303: print() should not be used. "
               "Please use LOG.[info|error|warning|exception|debug]. "
               "If print() must be used, use '# noqa' to skip this check.")
        yield(0, msg)


@core.flake8ext
def check_timeutils_strtime(logical_line):
    msg = ("C306: Found timeutils.strtime(). "
           "Please use datetime.datetime.isoformat() or datetime.strftime()")
    if 'timeutils.strtime' in logical_line:
        yield(0, msg)


@core.flake8ext
def dict_constructor_with_list_copy(logical_line):
    msg = ("N336: Must use a dict comprehension instead of a dict constructor "
           "with a sequence of key-value pairs.")
    if dict_constructor_with_list_copy_re.match(logical_line):
        yield (0, msg)


@core.flake8ext
def check_timeutils_isotime(logical_line):
    msg = ("C308: Found timeutils.isotime(). "
           "Please use datetime.datetime.isoformat()")
    if 'timeutils.isotime' in logical_line:
        yield(0, msg)


@core.flake8ext
def no_test_log(logical_line, filename, noqa):
    if ('cinder/tests' not in filename or noqa):
        return
    msg = "C309: Unit tests should not perform logging."
    if logging_instance.match(logical_line):
        yield (0, msg)


@core.flake8ext
def validate_assertTrue(logical_line, filename):
    # Note: a comparable check cannot be implemented for
    # assertFalse(), because assertFalse(None) passes.
    # Therefore, assertEqual(False, value) is required to
    # have the strongest test.
    if 'cinder/tests/unit' not in filename:
        return

    if re.match(assert_True, logical_line):
        msg = ("C313: Unit tests should use assertTrue(value) instead"
               " of using assertEqual(True, value).")
        yield(0, msg)


third_party_mock = re.compile("^import.mock")
from_third_party_mock = re.compile("^from.mock.import")


@core.flake8ext
def no_third_party_mock(logical_line):
    # We should only use unittest.mock, not the third party mock library that
    # was needed for py2 support.
    if (re.match(third_party_mock, logical_line) or
            re.match(from_third_party_mock, logical_line)):
        msg = ('C337: Unit tests should use the standard library "mock" '
               'module, not the third party mock lib.')
        yield(0, msg)



# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class StoropsException(Exception):
    message = 'Storops Error.'


class VNXException(StoropsException):
    message = "VNX Error."


class VNXStorageGroupError(VNXException):
    pass


class VNXAttachAluError(VNXException):
    pass


class VNXAluAlreadyAttachedError(VNXAttachAluError):
    message = (
        'LUN already exists in the specified storage group',
        'Requested LUN has already been added to this Storage Group')


class VNXDetachAluError(VNXStorageGroupError):
    pass


class VNXDetachAluNotFoundError(VNXDetachAluError):
    message = 'No such Host LUN in this Storage Group'


class VNXCreateStorageGroupError(VNXStorageGroupError):
    pass


class VNXStorageGroupNameInUseError(VNXCreateStorageGroupError):
    message = 'Storage Group name already in use'


class VNXNoHluAvailableError(VNXStorageGroupError):
    pass


class VNXMigrationError(VNXException):
    pass


class VNXLunNotMigratingError(VNXException):
    pass


class VNXLunSyncCompletedError(VNXMigrationError):
    error_code = 0x714a8021


class VNXTargetNotReadyError(VNXMigrationError):
    message = 'The destination LUN is not available for migration'


class VNXSnapError(VNXException):
    pass


class VNXDeleteAttachedSnapError(VNXSnapError):
    error_code = 0x716d8003


class VNXCreateSnapError(VNXException):
    message = 'Cannot create the snapshot.'


class VNXAttachSnapError(VNXSnapError):
    message = 'Cannot attach the snapshot.'


class VNXDetachSnapError(VNXSnapError):
    message = 'Cannot detach the snapshot.'


class VNXSnapAlreadyMountedError(VNXSnapError):
    error_code = 0x716d8055


class VNXSnapNameInUseError(VNXSnapError):
    error_code = 0x716d8005


class VNXSnapNotExistsError(VNXSnapError):
    message = 'The specified snapshot does not exist.'


class VNXLunError(VNXException):
    pass


class VNXCreateLunError(VNXLunError):
    pass


class VNXLunNameInUseError(VNXCreateLunError):
    error_code = 0x712d8d04


class VNXLunExtendError(VNXLunError):
    pass


class VNXLunExpandSizeError(VNXLunExtendError):
    error_code = 0x712d8e04


class VNXLunPreparingError(VNXLunError):
    error_code = 0x712d8e0e


class VNXLunNotFoundError(VNXLunError):
    message = 'Could not retrieve the specified (pool lun).'


class VNXDeleteLunError(VNXLunError):
    pass


class VNXLunUsedByFeatureError(VNXLunError):
    pass


class VNXCompressionError(VNXLunError):
    pass


class VNXCompressionAlreadyEnabledError(VNXCompressionError):
    message = 'Compression on the specified LUN is already turned on.'


class VNXConsistencyGroupError(VNXException):
    pass


class VNXCreateConsistencyGroupError(VNXConsistencyGroupError):
    pass


class VNXConsistencyGroupNameInUseError(VNXCreateConsistencyGroupError):
    error_code = 0x716d8021


class VNXConsistencyGroupNotFoundError(VNXConsistencyGroupError):
    message = 'Cannot find the consistency group'


class VNXPingNodeError(VNXException):
    pass


class VNXMirrorException(VNXException):
    pass


class VNXMirrorNameInUseError(VNXMirrorException):
    message = 'Mirror name already in use'


class VNXMirrorPromotePrimaryError(VNXMirrorException):
    message = 'Cannot remove or promote a primary image.'


class VNXMirrorNotFoundError(VNXMirrorException):
    message = 'Mirror not found'


class VNXMirrorGroupNameInUseError(VNXMirrorException):
    message = 'Mirror Group name already in use'


class VNXMirrorGroupNotFoundError(VNXMirrorException):
    message = 'Unable to locate the specified group'


class VNXMirrorGroupAlreadyMemberError(VNXMirrorException):
    message = 'The mirror is already a member of a group'


class VNXMirrorGroupMirrorNotMemberError(VNXMirrorException):
    message = 'The specified mirror is not a member of the group'


class VNXMirrorGroupAlreadyPromotedError(VNXMirrorException):
    message = 'The Consistency Group has no secondary images to promote'



#   Copyright 2014 IBM Corp.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from oslo_serialization import jsonutils
from six.moves import http_client
import webob

from cinder import context
from cinder import db
from cinder import objects
from cinder.objects import fields
from cinder.tests.unit.api import fakes
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test
from cinder.tests.unit import utils


class VolumeUnmanageTest(test.TestCase):
    """Test cases for cinder/api/contrib/volume_unmanage.py

    The API extension adds an action to volumes, "os-unmanage", which will
    effectively issue a delete operation on the volume, but with a flag set
    that means that a different method will be invoked on the driver, so that
    the volume is not actually deleted in the storage backend.

    In this set of test cases, we are ensuring that the code correctly parses
    the request structure and raises the correct exceptions when things are not
    right, and calls down into cinder.volume.api.API.delete with the correct
    arguments.
    """

    def setUp(self):
        super(VolumeUnmanageTest, self).setUp()
        self.ctxt = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, True)

        api = fakes.router.APIRouter()
        self.app = fakes.urlmap.URLMap()
        self.app['/v2'] = api

    def _get_resp(self, volume_id):
        """Helper to build an os-unmanage req for the specified volume_id."""
        req = webob.Request.blank('/v2/%s/volumes/%s/action' %
                                  (self.ctxt.project_id, volume_id))
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.environ['cinder.context'] = self.ctxt
        body = {'os-unmanage': ''}
        req.body = jsonutils.dump_as_bytes(body)
        res = req.get_response(self.app)
        return res

    @mock.patch('cinder.volume.rpcapi.VolumeAPI.delete_volume')
    def test_unmanage_volume_ok(self, mock_rpcapi):
        """Return success for valid and unattached volume."""
        vol = utils.create_volume(self.ctxt)
        res = self._get_resp(vol.id)
        self.assertEqual(http_client.ACCEPTED, res.status_int, res)

        mock_rpcapi.assert_called_once_with(self.ctxt, mock.ANY, True, False)
        vol = objects.volume.Volume.get_by_id(self.ctxt, vol.id)
        self.assertEqual('unmanaging', vol.status)
        db.volume_destroy(self.ctxt, vol.id)

    def test_unmanage_volume_bad_volume_id(self):
        """Return 404 if the volume does not exist."""
        res = self._get_resp(fake.WILL_NOT_BE_FOUND_ID)
        self.assertEqual(http_client.NOT_FOUND, res.status_int, res)

    def test_unmanage_volume_attached(self):
        """Return 400 if the volume exists but is attached."""
        vol = utils.create_volume(
            self.ctxt, status='in-use',
            attach_status=fields.VolumeAttachStatus.ATTACHED)
        res = self._get_resp(vol.id)
        self.assertEqual(http_client.BAD_REQUEST, res.status_int, res)
        db.volume_destroy(self.ctxt, vol.id)

    def test_unmanage_volume_with_snapshots(self):
        """Return 400 if the volume exists but has snapshots."""
        vol = utils.create_volume(self.ctxt)
        snap = utils.create_snapshot(self.ctxt, vol.id)
        res = self._get_resp(vol.id)
        self.assertEqual(http_client.BAD_REQUEST, res.status_int, res)
        db.volume_destroy(self.ctxt, vol.id)
        db.snapshot_destroy(self.ctxt, snap.id)

    def test_unmanage_encrypted_volume_denied(self):
        vol = utils.create_volume(
            self.ctxt,
            encryption_key_id='7a98391f-6619-46af-bd00-5862a3f7f1bd')
        res = self._get_resp(vol.id)
        self.assertEqual(http_client.BAD_REQUEST, res.status_int, res)
        db.volume_destroy(self.ctxt, vol.id)



# Copyright 2016 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for OpenStack Cinder volume driver."""

from unittest import mock
from unittest.mock import patch

from oslo_utils import units

from cinder import context
from cinder import db
from cinder.tests.unit import test
from cinder.volume import configuration as conf
from cinder.volume.drivers.nexenta import iscsi
from cinder.volume.drivers.nexenta import jsonrpc
from cinder.volume.drivers.nexenta import nfs
from cinder.volume.drivers.nexenta import utils


class TestNexentaISCSIDriver(test.TestCase):
    TEST_VOLUME_NAME = 'volume1'
    TEST_VOLUME_NAME2 = 'volume2'
    TEST_VOLUME_NAME3 = 'volume3'
    TEST_SNAPSHOT_NAME = 'snapshot1'
    TEST_VOLUME_REF = {
        'name': TEST_VOLUME_NAME,
        'size': 1,
        'id': '1',
        'status': 'available'
    }
    TEST_VOLUME_REF2 = {
        'name': TEST_VOLUME_NAME2,
        'size': 1,
        'id': '2',
        'status': 'in-use'
    }
    TEST_VOLUME_REF3 = {
        'name': TEST_VOLUME_NAME3,
        'size': 3,
        'id': '3',
        'status': 'in-use'
    }
    TEST_SNAPSHOT_REF = {
        'name': TEST_SNAPSHOT_NAME,
        'volume_name': TEST_VOLUME_NAME,
        'volume_size': 1,
    }

    def __init__(self, method):
        super(TestNexentaISCSIDriver, self).__init__(method)

    def setUp(self):
        super(TestNexentaISCSIDriver, self).setUp()
        self.cfg = mock.Mock(spec=conf.Configuration)
        self.ctxt = context.get_admin_context()
        self.cfg.nexenta_dataset_description = ''
        self.cfg.nexenta_host = '1.1.1.1'
        self.cfg.nexenta_user = 'admin'
        self.cfg.nexenta_password = 'nexenta'
        self.cfg.nexenta_volume = 'cinder'
        self.cfg.nexenta_rest_port = 2000
        self.cfg.nexenta_rest_protocol = 'http'
        self.cfg.nexenta_iscsi_target_portal_port = 3260
        self.cfg.nexenta_target_prefix = 'iqn:'
        self.cfg.nexenta_target_group_prefix = 'cinder/'
        self.cfg.nexenta_blocksize = '8K'
        self.cfg.nexenta_sparse = True
        self.cfg.nexenta_dataset_compression = 'on'
        self.cfg.nexenta_dataset_dedup = 'off'
        self.cfg.nexenta_rrmgr_compression = 1
        self.cfg.nexenta_rrmgr_tcp_buf_size = 1024
        self.cfg.nexenta_rrmgr_connections = 2
        self.cfg.reserved_percentage = 20
        self.nms_mock = mock.Mock()
        for mod in ['volume', 'zvol', 'iscsitarget', 'appliance',
                    'stmf', 'scsidisk', 'snapshot']:
            setattr(self.nms_mock, mod, mock.Mock())
        self.mock_object(jsonrpc, 'NexentaJSONProxy',
                         return_value=self.nms_mock)
        self.drv = iscsi.NexentaISCSIDriver(
            configuration=self.cfg)
        self.drv.db = db
        self.drv.do_setup(self.ctxt)

    def test_check_do_setup(self):
        self.assertEqual('http', self.drv.nms_protocol)

    def test_check_for_setup_error(self):
        self.nms_mock.volume.object_exists.return_value = False
        self.assertRaises(LookupError, self.drv.check_for_setup_error)

    def test_local_path(self):
        self.assertRaises(NotImplementedError, self.drv.local_path, '')

    def test_create_volume(self):
        self.drv.create_volume(self.TEST_VOLUME_REF)
        self.nms_mock.zvol.create.assert_called_with(
            'cinder/%s' % self.TEST_VOLUME_REF['name'], '1G',
            self.cfg.nexenta_blocksize, self.cfg.nexenta_sparse)

    def test_delete_volume(self):
        self.drv._collect_garbage = lambda vol: vol
        self.nms_mock.zvol.get_child_props.return_value = (
            {'origin': 'cinder/volume0@snapshot'})
        self.drv.delete_volume(self.TEST_VOLUME_REF)
        self.nms_mock.zvol.get_child_props.assert_called_with(
            'cinder/volume1', 'origin')
        self.nms_mock.zvol.destroy.assert_called_with(
            'cinder/volume1', '')

        self.nms_mock.zvol.get_child_props.assert_called_with(
            'cinder/volume1', 'origin')
        self.nms_mock.zvol.destroy.assert_called_with('cinder/volume1', '')
        self.drv.delete_volume(self.TEST_VOLUME_REF)

        self.nms_mock.zvol.get_child_props.assert_called_with(
            'cinder/volume1', 'origin')

    def test_create_cloned_volume(self):
        vol = self.TEST_VOLUME_REF2
        src_vref = self.TEST_VOLUME_REF
        snapshot = {
            'volume_name': src_vref['name'],
            'name': 'cinder-clone-snapshot-%s' % vol['id'],
        }
        self.drv.create_cloned_volume(vol, src_vref)
        self.nms_mock.zvol.create_snapshot.assert_called_with(
            'cinder/%s' % src_vref['name'], snapshot['name'], '')
        self.nms_mock.zvol.clone.assert_called_with(
            'cinder/%s@%s' % (src_vref['name'], snapshot['name']),
            'cinder/%s' % vol['name'])

    def test_migrate_volume(self):
        self.drv._collect_garbage = lambda vol: vol
        volume = self.TEST_VOLUME_REF
        host = {
            'capabilities': {
                'vendor_name': 'Nexenta',
                'location_info': 'NexentaISCSIDriver:1.1.1.1:cinder',
                'free_capacity_gb': 1,
                'iscsi_target_portal_port': 3260,
                'nms_url': 'http://admin:password@1.1.1.1:2000'
            }
        }
        snapshot = {
            'volume_name': volume['name'],
            'name': 'cinder-migrate-snapshot-%s' % volume['id'],
        }
        volume_name = 'cinder/%s' % volume['name']

        self.nms_mock.appliance.ssh_list_bindings.return_value = (
            {'0': [True, True, True, '1.1.1.1']})
        self.nms_mock.zvol.get_child_props.return_value = None

        self.drv.migrate_volume(None, volume, host)
        self.nms_mock.zvol.create_snapshot.assert_called_with(
            'cinder/%s' % volume['name'], snapshot['name'], '')

        src = '%(volume)s/%(zvol)s@%(snapshot)s' % {
            'volume': 'cinder',
            'zvol': volume['name'],
            'snapshot': snapshot['name']
        }
        dst = '1.1.1.1:cinder'
        cmd = ' '.join(['rrmgr -s zfs -c 1 -q -e -w 1024 -n 2', src, dst])

        self.nms_mock.appliance.execute.assert_called_with(cmd)

        snapshot_name = 'cinder/%(volume)s@%(snapshot)s' % {
            'volume': volume['name'],
            'snapshot': snapshot['name']
        }
        self.nms_mock.snapshot.destroy.assert_called_with(snapshot_name, '')
        self.nms_mock.zvol.destroy.assert_called_with(volume_name, '')
        self.nms_mock.snapshot.destroy.assert_called_with(
            'cinder/%(volume)s@%(snapshot)s' % {
                'volume': volume['name'],
                'snapshot': snapshot['name']
            }, '')

    def test_create_snapshot(self):
        self.drv.create_snapshot(self.TEST_SNAPSHOT_REF)
        self.nms_mock.zvol.create_snapshot.assert_called_with(
            'cinder/volume1', 'snapshot1', '')

    def test_create_volume_from_snapshot(self):
        self._create_volume_db_entry()
        self.drv.create_volume_from_snapshot(self.TEST_VOLUME_REF3,
                                             self.TEST_SNAPSHOT_REF)
        self.nms_mock.zvol.clone.assert_called_with(
            'cinder/volume1@snapshot1', 'cinder/volume3')
        self.nms_mock.zvol.set_child_prop.assert_called_with(
            'cinder/volume3', 'volsize', '3G')

    def test_delete_snapshot(self):
        self._create_volume_db_entry()
        self.drv._collect_garbage = lambda vol: vol
        self.drv.delete_snapshot(self.TEST_SNAPSHOT_REF)
        self.nms_mock.snapshot.destroy.assert_called_with(
            'cinder/volume1@snapshot1', '')

        # Check that exception not raised if snapshot does not exist
        self.drv.delete_snapshot(self.TEST_SNAPSHOT_REF)
        self.nms_mock.snapshot.destroy.side_effect = (
            utils.NexentaException('does not exist'))
        self.nms_mock.snapshot.destroy.assert_called_with(
            'cinder/volume1@snapshot1', '')

    def _mock_all_export_methods(self, fail=False):
        self.assertTrue(self.nms_mock.stmf.list_targets.called)
        self.nms_mock.iscsitarget.create_target.assert_called_with(
            {'target_name': 'iqn:1.1.1.1-0'})
        self.nms_mock.stmf.list_targetgroups()
        zvol_name = 'cinder/volume1'
        self.nms_mock.stmf.create_targetgroup.assert_called_with(
            'cinder/1.1.1.1-0')
        self.nms_mock.stmf.list_targetgroup_members.assert_called_with(
            'cinder/1.1.1.1-0')
        self.nms_mock.scsidisk.lu_exists.assert_called_with(zvol_name)
        self.nms_mock.scsidisk.create_lu.assert_called_with(zvol_name, {})

    def _stub_all_export_methods(self):
        self.nms_mock.scsidisk.lu_exists.return_value = False
        self.nms_mock.scsidisk.lu_shared.side_effect = (
            utils.NexentaException(['does not exist for zvol']))
        self.nms_mock.scsidisk.create_lu.return_value = {'lun': 0}
        self.nms_mock.stmf.list_targets.return_value = []
        self.nms_mock.stmf.list_targetgroups.return_value = []
        self.nms_mock.stmf.list_targetgroup_members.return_value = []
        self.nms_mock._get_target_name.return_value = ['iqn:1.1.1.1-0']
        self.nms_mock.iscsitarget.create_targetgroup.return_value = ({
            'target_name': 'cinder/1.1.1.1-0'})
        self.nms_mock.scsidisk.add_lun_mapping_entry.return_value = {'lun': 0}

    def test_create_export(self):
        self._stub_all_export_methods()
        retval = self.drv.create_export({}, self.TEST_VOLUME_REF, None)
        self._mock_all_export_methods()
        location = '%(host)s:%(port)s,1 %(name)s %(lun)s' % {
            'host': self.cfg.nexenta_host,
            'port': self.cfg.nexenta_iscsi_target_portal_port,
            'name': 'iqn:1.1.1.1-0',
            'lun': '0'
        }
        self.assertEqual({'provider_location': location}, retval)

    def test_ensure_export(self):
        self._stub_all_export_methods()
        self.drv.ensure_export({}, self.TEST_VOLUME_REF)
        self._mock_all_export_methods()

    def test_remove_export(self):
        self.nms_mock.stmf.list_targets.return_value = ['iqn:1.1.1.1-0']
        self.nms_mock.stmf.list_targetgroups.return_value = (
            ['cinder/1.1.1.1-0'])
        self.nms_mock.stmf.list_targetgroup_members.return_value = (
            ['iqn:1.1.1.1-0'])
        self.drv.remove_export({}, self.TEST_VOLUME_REF)
        self.assertTrue(self.nms_mock.stmf.list_targets.called)
        self.assertTrue(self.nms_mock.stmf.list_targetgroups.called)
        self.nms_mock.scsidisk.delete_lu.assert_called_with('cinder/volume1')

    def test_get_volume_stats(self):
        stats = {'size': '5368709120G',
                 'used': '5368709120G',
                 'available': '5368709120G',
                 'health': 'ONLINE'}
        self.nms_mock.volume.get_child_props.return_value = stats
        stats = self.drv.get_volume_stats(True)
        self.assertEqual('iSCSI', stats['storage_protocol'])
        self.assertEqual(5368709120.0, stats['total_capacity_gb'])
        self.assertEqual(5368709120.0, stats['free_capacity_gb'])
        self.assertEqual(20, stats['reserved_percentage'])
        self.assertFalse(stats['QoS_support'])

    def test_collect_garbage__snapshot(self):
        name = 'cinder/v1@s1'
        self.drv._mark_as_garbage(name)
        self.nms_mock.zvol.get_child_props.return_value = None
        self.drv._collect_garbage(name)
        self.nms_mock.snapshot.destroy.assert_called_with(name, '')
        self.assertNotIn(name, self.drv._needless_objects)

    def test_collect_garbage__volume(self):
        name = 'cinder/v1'
        self.drv._mark_as_garbage(name)
        self.nms_mock.zvol.get_child_props.return_value = None
        self.drv._collect_garbage(name)
        self.nms_mock.zvol.destroy.assert_called_with(name, '')
        self.assertNotIn(name, self.drv._needless_objects)

    def _create_volume_db_entry(self):
        vol = {
            'id': '1',
            'size': 1,
            'status': 'available',
            'provider_location': self.TEST_VOLUME_NAME,
            'volume_type_id': self.vt['id']
        }
        return db.volume_create(self.ctxt, vol)['id']


class TestNexentaNfsDriver(test.TestCase):
    TEST_VOLUME_NAME = 'volume1'
    TEST_VOLUME_NAME2 = 'volume2'
    TEST_VOLUME_NAME3 = 'volume3'
    TEST_SNAPSHOT_NAME = 'snapshot1'
    TEST_VOLUME_REF = {
        'name': TEST_VOLUME_NAME,
        'size': 1,
        'id': '1',
        'status': 'available'
    }
    TEST_VOLUME_REF2 = {
        'name': TEST_VOLUME_NAME2,
        'size': 2,
        'id': '2',
        'status': 'in-use'
    }
    TEST_VOLUME_REF3 = {
        'name': TEST_VOLUME_NAME2,
        'id': '2',
        'status': 'in-use'
    }
    TEST_SNAPSHOT_REF = {
        'name': TEST_SNAPSHOT_NAME,
        'volume_name': TEST_VOLUME_NAME,
        'volume_size': 1,
        'volume_id': 1
    }

    TEST_EXPORT1 = 'host1:/volumes/stack/share'
    TEST_NMS1 = 'http://admin:nexenta@host1:2000'

    TEST_EXPORT2 = 'host2:/volumes/stack/share'
    TEST_NMS2 = 'http://admin:nexenta@host2:2000'

    TEST_EXPORT2_OPTIONS = '-o intr'

    TEST_FILE_NAME = 'test.txt'
    TEST_SHARES_CONFIG_FILE = '/etc/cinder/nexenta-shares.conf'

    TEST_SHARE_SVC = 'svc:/network/nfs/server:default'

    TEST_SHARE_OPTS = {
        'read_only': '',
        'read_write': '*',
        'recursive': 'true',
        'anonymous_rw': 'true',
        'extra_options': 'anon=0',
        'root': 'nobody'
    }

    def _create_volume_db_entry(self):
        vol = {
            'id': '1',
            'size': 1,
            'status': 'available',
            'provider_location': self.TEST_EXPORT1,
            'volume_type_id': self.vt['id']
        }
        self.drv.share2nms = {self.TEST_EXPORT1: self.nms_mock}
        return db.volume_create(self.ctxt, vol)['id']

    def setUp(self):
        super(TestNexentaNfsDriver, self).setUp()
        self.ctxt = context.get_admin_context()
        self.cfg = mock.Mock(spec=conf.Configuration)
        self.cfg.nexenta_dataset_description = ''
        self.cfg.nexenta_shares_config = None
        self.cfg.nexenta_mount_point_base = '$state_path/mnt'
        self.cfg.nexenta_sparsed_volumes = True
        self.cfg.nexenta_dataset_compression = 'on'
        self.cfg.nexenta_dataset_dedup = 'off'
        self.cfg.nexenta_rrmgr_compression = 1
        self.cfg.nexenta_rrmgr_tcp_buf_size = 1024
        self.cfg.nexenta_rrmgr_connections = 2
        self.cfg.nfs_mount_point_base = '/mnt/test'
        self.cfg.nfs_mount_options = None
        self.cfg.nas_mount_options = None
        self.cfg.nexenta_nms_cache_volroot = False
        self.cfg.nfs_mount_attempts = 3
        self.cfg.reserved_percentage = 20
        self.cfg.max_over_subscription_ratio = 20.0
        self.nms_mock = mock.Mock()
        for mod in ('appliance', 'folder', 'server', 'volume', 'netstorsvc',
                    'snapshot', 'netsvc'):
            setattr(self.nms_mock, mod, mock.Mock())
        self.nms_mock.__hash__ = lambda *_, **__: 1
        self.mock_object(jsonrpc, 'NexentaJSONProxy',
                         return_value=self.nms_mock)
        self.drv = nfs.NexentaNfsDriver(configuration=self.cfg)
        self.drv.shares = {}
        self.drv.share2nms = {}

    def test_check_for_setup_error(self):
        self.drv.share2nms = {
            'host1:/volumes/stack/share': self.nms_mock
        }

        self.nms_mock.server.get_prop.return_value = '/volumes'
        self.nms_mock.volume.object_exists.return_value = True
        self.nms_mock.folder.object_exists.return_value = True
        share_opts = {
            'read_write': '*',
            'read_only': '',
            'root': 'nobody',
            'extra_options': 'anon=0',
            'recursive': 'true',
            'anonymous_rw': 'true',
        }
        self.nms_mock.netstorsvc.get_shared_folders.return_value = ''
        self.nms_mock.folder.get_child_props.return_value = {
            'available': 1, 'used': 1}
        self.drv.check_for_setup_error()
        self.nms_mock.netstorsvc.share_folder.assert_called_with(
            'svc:/network/nfs/server:default', 'stack/share', share_opts)

        self.nms_mock.server.get_prop.return_value = '/volumes'
        self.nms_mock.volume.object_exists.return_value = False

        self.assertRaises(LookupError, self.drv.check_for_setup_error)

        self.nms_mock.server.get_prop.return_value = '/volumes'
        self.nms_mock.volume.object_exists.return_value = True
        self.nms_mock.folder.object_exists.return_value = False

        self.assertRaises(LookupError, self.drv.check_for_setup_error)

    def test_initialize_connection(self):
        self.drv.shares = {
            self.TEST_EXPORT1: None
        }
        volume = {
            'provider_location': self.TEST_EXPORT1,
            'name': 'volume'
        }
        result = self.drv.initialize_connection(volume, None)
        self.assertEqual('%s/volume' % self.TEST_EXPORT1,
                         result['data']['export'])

    def test_do_create_volume(self):
        volume = {
            'provider_location': self.TEST_EXPORT1,
            'size': 1,
            'name': 'volume-1'
        }
        self.drv.shares = {self.TEST_EXPORT1: None}
        self.drv.share2nms = {self.TEST_EXPORT1: self.nms_mock}

        compression = self.cfg.nexenta_dataset_compression
        self.nms_mock.folder.get_child_props.return_value = {
            'available': 1, 'used': 1}
        self.nms_mock.server.get_prop.return_value = '/volumes'
        self.nms_mock.netsvc.get_confopts('svc:/network/nfs/server:default',
                                          'configure').AndReturn({
                                              'nfs_server_versmax': {
                                                  'current': u'3'}})
        self.nms_mock.netsvc.get_confopts.return_value = {
            'nfs_server_versmax': {'current': 4}}
        self.nms_mock._ensure_share_mounted.return_value = True
        self.drv._do_create_volume(volume)
        self.nms_mock.folder.create_with_props.assert_called_with(
            'stack', 'share/volume-1', {'compression': compression})
        self.nms_mock.netstorsvc.share_folder.assert_called_with(
            self.TEST_SHARE_SVC, 'stack/share/volume-1', self.TEST_SHARE_OPTS)
        mock_chmod = self.nms_mock.appliance.execute
        mock_chmod.assert_called_with(
            'chmod ugo+rw /volumes/stack/share/volume-1/volume')
        mock_truncate = self.nms_mock.appliance.execute
        mock_truncate.side_effect = utils.NexentaException(
            'fake_exception')
        self.nms_mock.server.get_prop.return_value = '/volumes'
        self.nms_mock.folder.get_child_props.return_value = {
            'available': 1, 'used': 1}
        self.assertRaises(utils.NexentaException,
                          self.drv._do_create_volume, volume)

    def test_create_sparsed_file(self):
        self.drv._create_sparsed_file(self.nms_mock, '/tmp/path', 1)
        self.nms_mock.appliance.execute.assert_called_with(
            'truncate --size 1G /tmp/path')

    def test_create_regular_file(self):
        self.drv._create_regular_file(self.nms_mock, '/tmp/path', 1)
        self.nms_mock.appliance.execute.assert_called_with(
            'dd if=/dev/zero of=/tmp/path bs=1M count=1024')

    @patch('cinder.volume.drivers.remotefs.'
           'RemoteFSDriver._ensure_shares_mounted')
    @patch('cinder.volume.drivers.nexenta.nfs.'
           'NexentaNfsDriver._get_volroot')
    @patch('cinder.volume.drivers.nexenta.nfs.'
           'NexentaNfsDriver._get_nfs_server_version')
    def test_create_larger_volume_from_snap(self, version, volroot, ensure):
        version.return_value = 4
        volroot.return_value = 'volroot'
        self._create_volume_db_entry()
        self.drv.create_volume_from_snapshot(self.TEST_VOLUME_REF2,
                                             self.TEST_SNAPSHOT_REF)
        self.nms_mock.appliance.execute.assert_called_with(
            'truncate --size 2G /volumes/stack/share/volume2/volume')

    @patch('cinder.volume.drivers.remotefs.'
           'RemoteFSDriver._ensure_shares_mounted')
    @patch('cinder.volume.drivers.nexenta.nfs.'
           'NexentaNfsDriver._get_volroot')
    @patch('cinder.volume.drivers.nexenta.nfs.'
           'NexentaNfsDriver._get_nfs_server_version')
    def test_create_volume_from_snapshot(self, version, volroot, ensure):
        version.return_value = 4
        volroot.return_value = 'volroot'
        self._create_volume_db_entry()
        self.drv.create_volume_from_snapshot(self.TEST_VOLUME_REF,
                                             self.TEST_SNAPSHOT_REF)
        self.nms_mock.appliance.execute.assert_not_called()

        self.drv.create_volume_from_snapshot(self.TEST_VOLUME_REF3,
                                             self.TEST_SNAPSHOT_REF)
        self.nms_mock.appliance.execute.assert_not_called()

    def test_set_rw_permissions_for_all(self):
        path = '/tmp/path'
        self.drv._set_rw_permissions_for_all(self.nms_mock, path)
        self.nms_mock.appliance.execute.assert_called_with(
            'chmod ugo+rw %s' % path)

    def test_local_path(self):
        volume = {'provider_location': self.TEST_EXPORT1, 'name': 'volume-1'}
        path = self.drv.local_path(volume)
        self.assertEqual(
            '$state_path/mnt/b3f660847a52b29ac330d8555e4ad669/volume-1/volume',
            path
        )

    def test_remote_path(self):
        volume = {'provider_location': self.TEST_EXPORT1, 'name': 'volume-1'}
        path = self.drv.remote_path(volume)
        self.assertEqual('/volumes/stack/share/volume-1/volume', path)

    def test_share_folder(self):
        self.drv._share_folder(self.nms_mock, 'stack', 'share/folder')
        path = 'stack/share/folder'
        self.nms_mock.netstorsvc.share_folder.assert_called_with(
            self.TEST_SHARE_SVC, path, self.TEST_SHARE_OPTS)

    def test_load_shares_config(self):
        self.drv.configuration.nfs_shares_config = (
            self.TEST_SHARES_CONFIG_FILE)

        config_data = [
            '%s  %s' % (self.TEST_EXPORT1, self.TEST_NMS1),
            '# %s   %s' % (self.TEST_EXPORT2, self.TEST_NMS2),
            '',
            '%s  %s %s' % (self.TEST_EXPORT2, self.TEST_NMS2,
                           self.TEST_EXPORT2_OPTIONS)
        ]

        with mock.patch.object(self.drv, '_read_config_file') as \
                mock_read_config_file:
            mock_read_config_file.return_value = config_data
            self.drv._load_shares_config(
                self.drv.configuration.nfs_shares_config)

            self.assertIn(self.TEST_EXPORT1, self.drv.shares)
            self.assertIn(self.TEST_EXPORT2, self.drv.shares)
            self.assertEqual(2, len(self.drv.shares))

            self.assertIn(self.TEST_EXPORT1, self.drv.share2nms)
            self.assertIn(self.TEST_EXPORT2, self.drv.share2nms)
            self.assertEqual(2, len(self.drv.share2nms.keys()))

            self.assertEqual(self.TEST_EXPORT2_OPTIONS,
                             self.drv.shares[self.TEST_EXPORT2])

    def test_get_capacity_info(self):
        self.drv.share2nms = {self.TEST_EXPORT1: self.nms_mock}
        self.nms_mock.server.get_prop.return_value = '/volumes'
        self.nms_mock.folder.get_child_props.return_value = {
            'available': '1G',
            'used': '2G'
        }
        total, free, allocated = self.drv._get_capacity_info(self.TEST_EXPORT1)

        self.assertEqual(3 * units.Gi, total)
        self.assertEqual(units.Gi, free)
        self.assertEqual(2 * units.Gi, allocated)

    def test_get_share_datasets(self):
        self.drv.share2nms = {self.TEST_EXPORT1: self.nms_mock}
        self.nms_mock.server.get_prop.return_value = '/volumes'
        volume_name, folder_name = (
            self.drv._get_share_datasets(self.TEST_EXPORT1))

        self.assertEqual('stack', volume_name)
        self.assertEqual('share', folder_name)

    def test_delete_snapshot(self):
        self.drv.share2nms = {self.TEST_EXPORT1: self.nms_mock}
        self._create_volume_db_entry()

        self.nms_mock.server.get_prop.return_value = '/volumes'
        self.drv.delete_snapshot({'volume_id': '1', 'name': 'snapshot1'})
        self.nms_mock.snapshot.destroy.assert_called_with(
            'stack/share/volume-1@snapshot1', '')

    @mock.patch('os_brick.remotefs.remotefs.'
                'RemoteFsClient._read_mounts')
    def test_delete_volume(self, list_mount_points):
        self.drv.share2nms = {self.TEST_EXPORT1: self.nms_mock}
        self._create_volume_db_entry()

        self.drv._ensure_share_mounted = lambda *_, **__: 0
        self.drv._execute = lambda *_, **__: 0

        self.nms_mock.server.get_prop.return_value = '/volumes'
        self.nms_mock.folder.get_child_props.return_value = {
            'available': 1, 'used': 1}
        self.drv.delete_volume({
            'id': '1',
            'name': 'volume-1',
            'provider_location': self.TEST_EXPORT1
        })
        self.nms_mock.folder.destroy.assert_called_with(
            'stack/share/volume-1', '-r')
        list_mount_points.assert_called_once()

        # Check that exception not raised if folder does not exist on
        # NexentaStor appliance.
        mock = self.nms_mock.folder.destroy
        mock.side_effect = utils.NexentaException('Folder does not exist')
        self.drv.delete_volume({
            'id': '1',
            'name': 'volume-1',
            'provider_location': self.TEST_EXPORT1
        })


class TestNexentaUtils(test.TestCase):

    def test_str2size(self):
        values_to_test = (
            # Test empty value
            (None, 0),
            ('', 0),
            ('0', 0),
            ('12', 12),
            # Test int values
            (10, 10),
            # Test bytes string
            ('1b', 1),
            ('1B', 1),
            ('1023b', 1023),
            ('0B', 0),
            # Test other units
            ('1M', units.Mi),
            ('1.0M', units.Mi),
        )

        for value, result in values_to_test:
            self.assertEqual(result, utils.str2size(value))

        # Invalid format value
        self.assertRaises(ValueError, utils.str2size, 'A')

    def test_str2gib_size(self):
        self.assertEqual(1, utils.str2gib_size('1024M'))
        self.assertEqual(300 * units.Mi // units.Gi,
                         utils.str2gib_size('300M'))
        self.assertEqual(1.2 * units.Ti // units.Gi,
                         utils.str2gib_size('1.2T'))
        self.assertRaises(ValueError, utils.str2gib_size, 'A')

    def test_parse_nms_url(self):
        urls = (
            ('http://192.168.1.1/', (False, 'http', 'admin', 'nexenta',
                                     '192.168.1.1', '2000', '/rest/nms/')),
            ('http://192.168.1.1:8080', (False, 'http', 'admin', 'nexenta',
                                         '192.168.1.1', '8080', '/rest/nms/')),
            ('https://root:password@192.168.1.1:8080',
             (False, 'https', 'root', 'password', '192.168.1.1', '8080',
              '/rest/nms/')),
        )
        for url, result in urls:
            self.assertEqual(result, utils.parse_nms_url(url))



# Copyright 2016 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.api import microversions as mv
from cinder.api.v2.views import volumes as views_v2


class ViewBuilder(views_v2.ViewBuilder):
    """Model a volumes API V3 response as a python dictionary."""

    _collection_name = "volumes"

    def quick_summary(self, volume_count, volume_size,
                      all_distinct_metadata=None):
        """View of volumes summary.

        It includes number of volumes, size of volumes and all distinct
        metadata of volumes.
        """
        summary = {
            'volume-summary': {
                'total_count': volume_count,
                'total_size': volume_size
            }
        }
        if all_distinct_metadata is not None:
            summary['volume-summary']['metadata'] = all_distinct_metadata
        return summary

    def detail(self, request, volume):
        """Detailed view of a single volume."""
        volume_ref = super(ViewBuilder, self).detail(request, volume)

        req_version = request.api_version_request
        # Add group_id if min version is greater than or equal to GROUP_VOLUME.
        if req_version.matches(mv.GROUP_VOLUME, None):
            volume_ref['volume']['group_id'] = volume.get('group_id')

        # Add provider_id if min version is greater than or equal to
        # VOLUME_DETAIL_PROVIDER_ID for admin.
        if (request.environ['cinder.context'].is_admin and
                req_version.matches(mv.VOLUME_DETAIL_PROVIDER_ID, None)):
            volume_ref['volume']['provider_id'] = volume.get('provider_id')

        if req_version.matches(
                mv.VOLUME_SHARED_TARGETS_AND_SERVICE_FIELDS, None):
            volume_ref['volume']['shared_targets'] = volume.get(
                'shared_targets', None)
            volume_ref['volume']['service_uuid'] = volume.get(
                'service_uuid', None)

        if (request.environ['cinder.context'].is_admin and req_version.matches(
                mv.VOLUME_CLUSTER_NAME, None)):
            volume_ref['volume']['cluster_name'] = volume.get(
                'cluster_name', None)

        return volume_ref

    def _list_view(self, func, request, volumes, volume_count,
                   coll_name=_collection_name):
        """Provide a view for a list of volumes.

        :param func: Function used to format the volume data
        :param request: API request
        :param volumes: List of volumes in dictionary format
        :param volume_count: Length of the original list of volumes
        :param coll_name: Name of collection, used to generate the next link
                          for a pagination query
        :returns: Volume data in dictionary format
        """
        volumes_list = [func(request, volume)['volume'] for volume in volumes]
        volumes_links = self._get_collection_links(request,
                                                   volumes,
                                                   coll_name,
                                                   volume_count)
        volumes_dict = {"volumes": volumes_list}

        if volumes_links:
            volumes_dict['volumes_links'] = volumes_links

        req_version = request.api_version_request
        if req_version.matches(
                mv.SUPPORT_COUNT_INFO, None) and volume_count is not None:
            volumes_dict['count'] = volume_count

        return volumes_dict



# Copyright 2014, eBay Inc.
# Copyright 2014, OpenStack Foundation
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_utils import uuidutils

from cinder.scheduler import filters
from cinder.volume import api as volume


class AffinityFilter(filters.BaseBackendFilter):
    def __init__(self):
        self.volume_api = volume.API()

    def _get_volumes(self, context, affinity_uuids, backend_state):
        filters = {'id': affinity_uuids, 'deleted': False}
        if backend_state.cluster_name:
            filters['cluster_name'] = backend_state.cluster_name
        else:
            filters['host'] = backend_state.host
        return self.volume_api.get_all(context, filters=filters)


class DifferentBackendFilter(AffinityFilter):
    """Schedule volume on a different back-end from a set of volumes."""

    def backend_passes(self, backend_state, filter_properties):
        context = filter_properties['context']
        scheduler_hints = filter_properties.get('scheduler_hints') or {}

        affinity_uuids = scheduler_hints.get('different_host', [])

        # scheduler hint verification: affinity_uuids can be a list of uuids
        # or single uuid.  The checks here is to make sure every single string
        # in the list looks like a uuid, otherwise, this filter will fail to
        # pass.  Note that the filter does *NOT* ignore string doesn't look
        # like a uuid, it is better to fail the request than serving it wrong.
        if isinstance(affinity_uuids, list):
            for uuid in affinity_uuids:
                if uuidutils.is_uuid_like(uuid):
                    continue
                else:
                    return False
        elif uuidutils.is_uuid_like(affinity_uuids):
            affinity_uuids = [affinity_uuids]
        else:
            # Not a list, not a string looks like uuid, don't pass it
            # to DB for query to avoid potential risk.
            return False

        if affinity_uuids:
            return not self._get_volumes(context, affinity_uuids,
                                         backend_state)
        # With no different_host key
        return True


class SameBackendFilter(AffinityFilter):
    """Schedule volume on the same back-end as another volume."""

    def backend_passes(self, backend_state, filter_properties):
        context = filter_properties['context']
        scheduler_hints = filter_properties.get('scheduler_hints') or {}

        affinity_uuids = scheduler_hints.get('same_host', [])

        # scheduler hint verification: affinity_uuids can be a list of uuids
        # or single uuid.  The checks here is to make sure every single string
        # in the list looks like a uuid, otherwise, this filter will fail to
        # pass.  Note that the filter does *NOT* ignore string doesn't look
        # like a uuid, it is better to fail the request than serving it wrong.
        if isinstance(affinity_uuids, list):
            for uuid in affinity_uuids:
                if uuidutils.is_uuid_like(uuid):
                    continue
                else:
                    return False
        elif uuidutils.is_uuid_like(affinity_uuids):
            affinity_uuids = [affinity_uuids]
        else:
            # Not a list, not a string looks like uuid, don't pass it
            # to DB for query to avoid potential risk.
            return False

        if affinity_uuids:
            return self._get_volumes(context, affinity_uuids, backend_state)

        # With no same_host key
        return True



# Copyright (c) 2017 Hedvig Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from unittest import mock

from cinder import context
from cinder import exception
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.volume import configuration as conf
from cinder.volume.drivers.hedvig import hedvig_cinder as hdvg
from cinder.volume import qos_specs
from cinder.volume import volume_types


def _fake_volume_type(*args, **kwargs):
    ctxt = context.get_admin_context()
    type_ref = volume_types.create(ctxt, "qos_extra_specs", {})
    qos_ref = qos_specs.create(ctxt, 'qos-specs', {})
    qos_specs.associate_qos_with_type(ctxt, qos_ref['id'],
                                      type_ref['id'])
    qos_type = volume_types.get_volume_type(ctxt, type_ref['id'])
    return qos_type


def _fake_volume(*args, **kwargs):
    qos_type = _fake_volume_type()
    return fake_volume.fake_volume_obj(context,
                                       name='hedvig',
                                       volume_type_id=qos_type['id'],
                                       volume_type=qos_type,
                                       volume_name='hedvig',
                                       display_name='hedvig',
                                       display_description='test volume',
                                       size=2)


class HedvigDriverTest(test.TestCase):

    def setUp(self):
        super(HedvigDriverTest, self).setUp()
        self.context = context.get_admin_context()
        self._create_fake_config()
        self.assertIsNone(self.driver.do_setup(self.ctxt))

    def _create_fake_config(self):
        self.configuration = mock.Mock(spec=conf.Configuration)
        self.configuration.san_ip = '1.0.0.1'
        self.configuration.san_login = 'dummy_user'
        self.configuration.san_password = 'dummy_password'
        self.configuration.san_clustername = 'dummy_cluster'
        self.configuration.san_is_local = False
        self.ctxt = context.get_admin_context()
        self.vol = fake_volume.fake_volume_obj(self.context)
        self.vol.volume_type = fake_volume.fake_volume_type_obj(self.context)
        self.snap = fake_snapshot.fake_snapshot_obj(self.context)
        self.snap.volume = self.vol
        self.driver = hdvg.HedvigISCSIDriver(configuration=self.configuration)

    @mock.patch('cinder.volume.drivers.hedvig.rest_client.RestClient'
                '.create_vdisk')
    def test_create_volume(self, *args, **keywargs):
        result = self.driver.create_volume(self.vol)
        self.assertIsNone(result)

    def test_create_volume_negative(self, *args, **keywargs):
        self.driver.hrs = exception.VolumeDriverException()
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.create_volume, self.vol)

    @mock.patch('cinder.volume.drivers.hedvig.rest_client.RestClient'
                '.delete_vdisk')
    def test_create_delete_volume(self, *args, **keywargs):
        result = self.driver.delete_volume(self.vol)
        self.assertIsNone(result)

    def test_create_delete_volume_negative(self, *args, **keywargs):
        self.driver.hrs = exception.VolumeDriverException()
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.delete_volume, self.vol)

    @mock.patch('cinder.volume.drivers.hedvig.rest_client.RestClient'
                '.resize_vdisk')
    def test_extend_volume(self, *args, **keywargs):
        self.assertIsNone(self.driver.extend_volume(self.vol, 10))

    def test_extend_volume_negative(self, *args, **keywargs):
        self.driver.hrs = exception.VolumeDriverException()
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.extend_volume, self.vol, 10)

    @mock.patch('cinder.volume.drivers.hedvig.rest_client.RestClient'
                '.resize_vdisk')
    def test_extend_volume_shrinking(self, *args, **keywargs):
        volume = _fake_volume()
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.extend_volume, volume, 1)

    @mock.patch('cinder.volume.drivers.hedvig.rest_client.RestClient'
                '.clone_vdisk')
    def test_create_cloned_volume(self, *args, **keywargs):
        result = self.driver.create_cloned_volume(self.vol, self.vol)
        self.assertIsNone(result)

    def test_create_cloned_volume_negative(self, *args, **keywargs):
        self.driver.hrs = exception.VolumeDriverException()
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.create_cloned_volume, self.vol, self.vol)

    @mock.patch('cinder.volume.drivers.hedvig.rest_client.RestClient'
                '.create_snapshot')
    def test_create_snapshot(self, *args, **keywargs):
        result = self.driver.create_snapshot(self.snap)
        self.assertIsNone(result)

    def test_create_snapshot_negative(self, *args, **keywargs):
        self.driver.hrs = exception.VolumeDriverException()
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.create_snapshot, self.snap)

    @mock.patch('cinder.volume.drivers.hedvig.rest_client.RestClient'
                '.delete_snapshot')
    def test_delete_snapshot(self, *args, **keywargs):
        result = self.driver.delete_snapshot(self.snap)
        self.assertIsNone(result)

    def test_delete_snapshot_negative(self, *args, **keywargs):
        self.driver.hrs = exception.VolumeDriverException()
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.delete_snapshot, self.snap)

    @mock.patch('cinder.volume.drivers.hedvig.rest_client.RestClient'
                '.clone_hedvig_snapshot')
    def test_create_volume_from_snapshot(self, *args, **keywargs):
        result = self.driver.create_volume_from_snapshot(self.vol, self.snap)
        self.assertIsNone(result)

    def test_create_volume_from_snapshot_negative(self, *args, **keywargs):
        self.driver.hrs = exception.VolumeDriverException()
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.create_volume_from_snapshot,
                          self.vol, self.snap)

    def test_do_setup(self):
        self.driver.do_setup(self.context)

    def test_do_setup_san_ip_negative(self):
        self.configuration.san_ip = None
        # check the driver for setup errors
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.do_setup, self.context)
        self.configuration.san_ip = "1.0.0.1"

    def test_do_setup_san_cluster_negative(self):
        self.configuration.san_clustername = None
        # check the driver for setup errors
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.do_setup, self.context)
        self.configuration.san_clustername = "dummy_cluster"

    def test_do_setup_san_login_negative(self):
        self.configuration.san_login = None
        # check the driver for setup errors
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.do_setup, self.context)
        self.configuration.san_login = "dummy_user"

    def test_do_setup_san_password_negative(self):
        self.configuration.san_password = None
        # check the driver for setup errors
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.do_setup, self.context)
        self.configuration.san_password = "dummy_password"

    @mock.patch('cinder.volume.drivers.hedvig.rest_client.RestClient'
                '.list_targets')
    def test_hedvig_lookup_tgt(self, *args, **keywargs):
        host = "hostname"
        result = self.driver.hedvig_lookup_tgt(host)
        self.assertIsNone(result)

    def test_hedvig_lookup_tgt_negative(self, *args, **keywargs):
        host = "hostname"
        self.driver.hrs = exception.VolumeDriverException()
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.hedvig_lookup_tgt, host)

    def test_hedvig_get_lun_negative(self, *args, **keywargs):
        host = "hostname"
        volname = "volume"
        self.driver.hrs = exception.VolumeDriverException()
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.hedvig_get_lun, host, volname)

    @mock.patch('cinder.volume.drivers.hedvig.rest_client.RestClient'
                '.get_iqn')
    def test_hedvig_get_iqn(self, *args, **keywargs):
        host = "hostname"
        result = self.driver.hedvig_get_iqn(host)
        self.assertIsNotNone(result)

    def test_hedvig_get_iqn_negative(self, *args, **keywargs):
        host = "hostname"
        self.driver.hrs = exception.VolumeDriverException()
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.hedvig_get_iqn, host)

    @mock.patch('cinder.volume.drivers.hedvig.rest_client.RestClient'
                '.list_targets')
    def test_terminate_connection_no_connector(self, *args, **keywargs):
        self.assertIsNone(self.driver.
                          terminate_connection(_fake_volume(), None))



# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
VNX Constants

This module includes re-declaration from storops which directly used
by driver in module scope. That's to say:
If a constant from storops is used in class level, function signature,
module level, a re-declaration is needed in this file to avoid some static
import error when storops is not installed.
"""

from oslo_utils import importutils

storops = importutils.try_import('storops')

if storops:
    from storops import exception as storops_ex
    VNXLunPreparingError = storops_ex.VNXLunPreparingError
    VNXTargetNotReadyError = storops_ex.VNXTargetNotReadyError
    MIGRATION_RATE_HIGH = storops.VNXMigrationRate.HIGH
    PROVISION_THICK = storops.VNXProvisionEnum.THICK
else:
    VNXLunPreparingError = None
    MIGRATION_RATE_HIGH = None
    PROVISION_THICK = None
    VNXTargetNotReadyError = None



# Copyright (c) 2018 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Sphinx extension to be able to extract driver config options from code."""

import importlib

from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils import statemachine as sm
from sphinx.util import logging
from oslo_config import cfg

LOG = logging.getLogger(__name__)


class ConfigTableDirective(rst.Directive):
    """Directive to extract config options into docs output."""

    option_spec = {
        'table-title': directives.unchanged,
        'config-target': directives.unchanged,
        'exclude-list': directives.unchanged,
        'exclusive-list': directives.unchanged,
    }

    has_content = True

    def _doc_module(self, module, filters, exclusive):
        """Extract config options from module."""
        options = []
        try:
            mod = importlib.import_module(module)
            for prop in dir(mod):
                # exclusive-list overrides others
                if exclusive and prop not in exclusive:
                    continue
                if prop in filters:
                    continue
                thing = getattr(mod, prop)
                if isinstance(thing, cfg.Opt) and thing not in options:
                    # An individual config option
                    options.append(thing)
                elif (isinstance(thing, list) and len(thing) > 0 and
                        isinstance(thing[0], cfg.Opt)):
                    # A list of config opts
                    options.extend(thing)
        except Exception as e:
            self.error('Unable to import {}: {}'.format(module, e))

        return options

    def _get_default(self, opt):
        """Tries to pick the best text to use as the default."""
        if hasattr(opt, 'sample_default') and opt.sample_default:
            return opt.sample_default

        if type(opt.default) == list:
            return "[%s]" % ', '.join(str(item) for item in opt.default)

        result = str(opt.default)
        if not result:
            result = '<>'
        return result

    def run(self):
        """Load and find config options to document."""
        modules = [c.strip() for c in self.content if c.strip()]

        if not modules:
            raise self.error('No modules provided to document.')

        env = self.state.document.settings.env
        app = env.app

        result = sm.ViewList()
        source = '<{}>'.format(__name__)

        target = self.options.get('config-target', '')
        title = self.options.get(
            'table-title',
            'Description of {} configuration options'.format(target))

        # See if there are option sets that need to be ignored
        exclude = self.options.get('exclude-list', '')
        exclude_list = [e.strip() for e in exclude.split(',') if e.strip()]

        exclusive = self.options.get('exclusive-list', '')
        exclusive_list = [e.strip() for e in exclusive.split(',') if e.strip()]

        result.append('.. _{}:'.format(title.replace(' ', '-')), source)
        result.append('', source)
        result.append('.. list-table:: {}'.format(title), source)
        result.append('   :header-rows: 1', source)
        result.append('   :class: config-ref-table', source)
        result.append('', source)
        result.append('   * - Configuration option = Default value', source)
        result.append('     - Description', source)

        options = []
        for module in modules:
            retval = self._doc_module(module, exclude_list, exclusive_list)
            if retval:
                options.extend(retval)
            else:
                LOG.info('[config-table] No options found in {}'.format(
                         module))

        # Get options sorted alphabetically but with deprecated options last
        list.sort(options, key=lambda opt: opt.name)
        list.sort(options, key=lambda opt: opt.deprecated_for_removal)

        for opt in options:
            result.append(
                '   * - ``{}`` = ``{}``'.format(
                    opt.name, self._get_default(opt)),
                source)
            result.append(
                '     - ({}) {}{}'.format(
                    opt.type, opt.help,
                    ' **DEPRECATED**' if opt.deprecated_for_removal else ''),
                source)

        node = nodes.section()
        node.document = self.state.document
        self.state.nested_parse(result, 0, node)
        return node.children


def setup(app):
    app.add_directive('config-table', ConfigTableDirective)



#    Copyright 2015 SimpliVity Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
import six
from sqlalchemy.orm import attributes

from cinder import db
from cinder import objects
from cinder.objects import fields
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_volume
from cinder.tests.unit import objects as test_objects


@ddt.ddt
class TestVolumeAttachment(test_objects.BaseObjectsTestCase):

    @mock.patch('cinder.db.sqlalchemy.api.volume_attachment_get')
    def test_get_by_id(self, volume_attachment_get):
        db_attachment = fake_volume.volume_attachment_db_obj()
        attachment_obj = fake_volume.volume_attachment_ovo(self.context)
        volume_attachment_get.return_value = db_attachment
        attachment = objects.VolumeAttachment.get_by_id(self.context,
                                                        fake.ATTACHMENT_ID)
        self._compare(self, attachment_obj, attachment)

    @mock.patch.object(objects.Volume, 'get_by_id')
    def test_lazy_load_volume(self, volume_get_mock):
        volume = objects.Volume(self.context, id=fake.VOLUME_ID)
        volume_get_mock.return_value = volume
        attach = objects.VolumeAttachment(self.context, id=fake.ATTACHMENT_ID,
                                          volume_id=volume.id)

        r = attach.volume
        self.assertEqual(volume, r)
        volume_get_mock.assert_called_once_with(self.context, volume.id)

    def test_from_db_object_no_volume(self):
        original_get = attributes.InstrumentedAttribute.__get__

        def my_get(get_self, instance, owner):
            self.assertNotEqual('volume', get_self.key)
            return original_get(get_self, instance, owner)

        # Volume field is not loaded
        attach = fake_volume.models.VolumeAttachment(id=fake.ATTACHMENT_ID,
                                                     volume_id=fake.VOLUME_ID)
        patch_str = 'sqlalchemy.orm.attributes.InstrumentedAttribute.__get__'
        with mock.patch(patch_str, side_effect=my_get):
            objects.VolumeAttachment._from_db_object(
                self.context, objects.VolumeAttachment(), attach)

    @mock.patch('cinder.db.volume_attachment_update')
    def test_save(self, volume_attachment_update):
        attachment = fake_volume.volume_attachment_ovo(self.context)
        attachment.attach_status = fields.VolumeAttachStatus.ATTACHING
        attachment.save()
        volume_attachment_update.assert_called_once_with(
            self.context, attachment.id,
            {'attach_status': fields.VolumeAttachStatus.ATTACHING})

    @mock.patch('cinder.db.sqlalchemy.api.volume_attachment_get')
    def test_refresh(self, attachment_get):
        db_attachment1 = fake_volume.volume_attachment_db_obj()
        attachment_obj1 = fake_volume.volume_attachment_ovo(self.context)
        db_attachment2 = fake_volume.volume_attachment_db_obj()
        db_attachment2.mountpoint = '/dev/sdc'
        attachment_obj2 = fake_volume.volume_attachment_ovo(
            self.context, mountpoint='/dev/sdc')

        # On the second volume_attachment_get, return the volume attachment
        # with an updated mountpoint
        attachment_get.side_effect = [db_attachment1, db_attachment2]
        attachment = objects.VolumeAttachment.get_by_id(self.context,
                                                        fake.ATTACHMENT_ID)
        self._compare(self, attachment_obj1, attachment)

        # mountpoint was updated, so a volume attachment refresh should have a
        # new value for that field
        attachment.refresh()
        self._compare(self, attachment_obj2, attachment)
        if six.PY3:
            call_bool = mock.call.__bool__()
        else:
            call_bool = mock.call.__nonzero__()
        attachment_get.assert_has_calls([mock.call(self.context,
                                                   fake.ATTACHMENT_ID),
                                         call_bool,
                                         mock.call(self.context,
                                                   fake.ATTACHMENT_ID)])

    @mock.patch('cinder.db.sqlalchemy.api.volume_attached')
    def test_volume_attached(self, volume_attached):
        attachment = fake_volume.volume_attachment_ovo(self.context)
        updated_values = {'mountpoint': '/dev/sda',
                          'attach_status': fields.VolumeAttachStatus.ATTACHED,
                          'instance_uuid': fake.INSTANCE_ID}
        volume_attached.return_value = (fake_volume.fake_db_volume(),
                                        updated_values)
        volume = attachment.finish_attach(fake.INSTANCE_ID,
                                          'fake_host',
                                          '/dev/sda',
                                          'rw')
        self.assertIsInstance(volume, objects.Volume)
        volume_attached.assert_called_once_with(mock.ANY,
                                                attachment.id,
                                                fake.INSTANCE_ID,
                                                'fake_host',
                                                '/dev/sda',
                                                'rw',
                                                True)
        self.assertEqual('/dev/sda', attachment.mountpoint)
        self.assertEqual(fake.INSTANCE_ID, attachment.instance_uuid)
        self.assertEqual(fields.VolumeAttachStatus.ATTACHED,
                         attachment.attach_status)

    @ddt.data('1.0', '1.1', '1.2', '1.3')
    def test_obj_make_compatible(self, version):
        connection_info = {'field': 'value'}
        connector = {'host': '127.0.0.1'}
        vol_attach = objects.VolumeAttachment(self.context,
                                              connection_info=connection_info,
                                              connector=connector)
        primitive = vol_attach.obj_to_primitive(version)
        converted_vol_attach = objects.VolumeAttachment.obj_from_primitive(
            primitive)
        if version == '1.3':
            self.assertEqual(connector, converted_vol_attach.connector)
        elif version == '1.2':
            self.assertEqual(connection_info,
                             converted_vol_attach.connection_info)
        else:
            self.assertNotIn('connector', converted_vol_attach)
            self.assertFalse(converted_vol_attach.obj_attr_is_set(
                'connection_info'))

    def test_migrate_attachment_specs(self):
        # Create an attachment.
        attachment = objects.VolumeAttachment(
            self.context, attach_status='attaching', volume_id=fake.VOLUME_ID)
        attachment.create()
        # Create some attachment_specs. Note that the key and value have to
        # be strings, the table doesn't handle things like a wwpns list
        # for a fibrechannel connector.
        connector = {'host': '127.0.0.1'}
        db.attachment_specs_update_or_create(
            self.context, attachment.id, connector)
        # Now get the volume attachment object from the database and make
        # sure the connector was migrated from the attachment_specs table
        # to the volume_attachment table and the specs were deleted.
        attachment = objects.VolumeAttachment.get_by_id(
            self.context, attachment.id)
        self.assertIn('connector', attachment)
        self.assertDictEqual(connector, attachment.connector)
        self.assertEqual(0, len(db.attachment_specs_get(
            self.context, attachment.id)))
        # Make sure we can store a fibrechannel type connector that has a wwpns
        # list value.
        connector['wwpns'] = ['21000024ff34c92d', '21000024ff34c92c']
        attachment.connector = connector
        attachment.save()
        # Get the object from the DB again and make sure the connector is
        # there.
        attachment = objects.VolumeAttachment.get_by_id(
            self.context, attachment.id)
        self.assertIn('connector', attachment)
        self.assertDictEqual(connector, attachment.connector)


class TestVolumeAttachmentList(test_objects.BaseObjectsTestCase):
    @mock.patch('cinder.db.volume_attachment_get_all_by_volume_id')
    def test_get_all_by_volume_id(self, get_used_by_volume_id):
        db_attachment = fake_volume.volume_attachment_db_obj()
        get_used_by_volume_id.return_value = [db_attachment]
        attachment_obj = fake_volume.volume_attachment_ovo(self.context)

        attachments = objects.VolumeAttachmentList.get_all_by_volume_id(
            self.context, mock.sentinel.volume_id)

        self.assertEqual(1, len(attachments))
        self._compare(self, attachment_obj, attachments[0])

    @mock.patch('cinder.db.volume_attachment_get_all_by_host')
    def test_get_all_by_host(self, get_by_host):
        db_attachment = fake_volume.volume_attachment_db_obj()
        attachment_obj = fake_volume.volume_attachment_ovo(self.context)
        get_by_host.return_value = [db_attachment]

        attachments = objects.VolumeAttachmentList.get_all_by_host(
            self.context, mock.sentinel.host)
        self.assertEqual(1, len(attachments))
        self._compare(self, attachment_obj, attachments[0])

    @mock.patch('cinder.db.volume_attachment_get_all_by_instance_uuid')
    def test_get_all_by_instance_uuid(self, get_by_instance_uuid):
        db_attachment = fake_volume.volume_attachment_db_obj()
        get_by_instance_uuid.return_value = [db_attachment]
        attachment_obj = fake_volume.volume_attachment_ovo(self.context)

        attachments = objects.VolumeAttachmentList.get_all_by_instance_uuid(
            self.context, mock.sentinel.uuid)
        self.assertEqual(1, len(attachments))
        self._compare(self, attachment_obj, attachments[0])



#
#    Copyright 2010 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.tests.unit import test
from cinder.tests.unit import utils as test_utils


class TestUtilsTestCase(test.TestCase):
    def test_get_test_admin_context(self):
        """get_test_admin_context's return value behaves like admin context."""
        ctxt = test_utils.get_test_admin_context()

        self.assertIsNone(ctxt.project_id)
        self.assertIsNone(ctxt.user_id)
        self.assertIsNone(ctxt.domain_id)
        self.assertIsNone(ctxt.project_domain_id)
        self.assertIsNone(ctxt.user_domain_id)
        self.assertIsNone(ctxt.project_name)
        self.assertIsNone(ctxt.remote_address)
        self.assertIsNone(ctxt.auth_token)
        self.assertIsNone(ctxt.quota_class)

        self.assertIsNotNone(ctxt.request_id)
        self.assertIsNotNone(ctxt.timestamp)

        self.assertEqual(['admin'], ctxt.roles)
        self.assertEqual([], ctxt.service_catalog)
        self.assertEqual('no', ctxt.read_deleted)

        self.assertTrue(ctxt.is_admin)



# Copyright 2015 Clinton Knight
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from cinder.api.openstack import versioned_method
from cinder.tests.unit import test


class VersionedMethodTestCase(test.TestCase):

    def test_str(self):
        args = ('fake_name', 'fake_min', 'fake_max')
        method = versioned_method.VersionedMethod(*(args + (False, None)))
        method_string = six.text_type(method)

        self.assertEqual('Version Method %s: min: %s, max: %s' % args,
                         method_string)

    def test_cmpkey(self):
        method = versioned_method.VersionedMethod(
            'fake_name', 'fake_start_version', 'fake_end_version', False,
            'fake_func')
        self.assertEqual('fake_start_version', method._cmpkey())



#    Copyright 2016 EMC Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from cinder import objects
from cinder.tests.unit import fake_constants as fake


def fake_db_group(**updates):
    db_group = {
        'id': fake.GROUP_ID,
        'name': 'group-1',
        'status': 'available',
        'user_id': fake.USER_ID,
        'project_id': fake.PROJECT_ID,
        'group_type_id': fake.GROUP_TYPE_ID,
        'group_snapshot_id': None,
        'source_group_id': None,
    }

    for name, field in objects.Group.fields.items():
        if name in db_group:
            continue
        if field.nullable:
            db_group[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_group[name] = field.default
        else:
            raise Exception('fake_db_group needs help with %s.' % name)

    if updates:
        db_group.update(updates)

    return db_group


def fake_db_group_type(**updates):
    db_group_type = {
        'id': fake.GROUP_TYPE_ID,
        'name': 'type-1',
        'description': 'A fake group type',
        'is_public': True,
        'projects': [],
        'group_specs': {},
    }

    for name, field in objects.GroupType.fields.items():
        if name in db_group_type:
            continue
        if field.nullable:
            db_group_type[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_group_type[name] = field.default
        else:
            raise Exception('fake_db_group_type needs help with %s.' % name)

    if updates:
        db_group_type.update(updates)

    return db_group_type


def fake_group_obj(context, **updates):
    return objects.Group._from_db_object(
        context, objects.Group(), fake_db_group(**updates))


def fake_group_type_obj(context, **updates):
    return objects.GroupType._from_db_object(
        context, objects.GroupType(), fake_db_group_type(**updates))
